import copy

import numpy as np


def Is_Point_Close_To_Id_In_Raster(prow, pcol, nrows, ncols, id, raster_array):
    """Check if the point is around grids with value equal to Id in raster_array

    Parameters
    ----------

    Returns:
    -------
    Isclose               : logical
       True  : close to grids with value euqal to id
       False : not close to grids with value to id
    """
    Isclose = False
    n_grids_eq_id = 0
    if prow != 0 and prow != nrows - 1 and pcol != 0 and pcol != ncols - 1:
        if raster_array[prow - 1, pcol + 1] == id:
            Isclose = True
            n_grids_eq_id = n_grids_eq_id + 1
        if raster_array[prow - 1, pcol - 1] == id:
            Isclose = True
            n_grids_eq_id = n_grids_eq_id + 1
        if raster_array[prow - 1, pcol] == id:
            Isclose = True
            n_grids_eq_id = n_grids_eq_id + 1
        if raster_array[prow, pcol + 1] == id:
            Isclose = True
            n_grids_eq_id = n_grids_eq_id + 1
        if raster_array[prow, pcol - 1] == id:
            Isclose = True
            n_grids_eq_id = n_grids_eq_id + 1
        if raster_array[prow + 1, pcol - 1] == id:
            Isclose = True
            n_grids_eq_id = n_grids_eq_id + 1
        if raster_array[prow + 1, pcol + 1] == id:
            Isclose = True
            n_grids_eq_id = n_grids_eq_id + 1
        if raster_array[prow + 1, pcol] == id:
            Isclose = True
            n_grids_eq_id = n_grids_eq_id + 1

    return Isclose, n_grids_eq_id


def CE_mcat4lake2(
    cat1, lake, fac, fdir, bsid, nrows, ncols, Pourpoints, noncnlake, str_array
):
    cat = copy.copy(cat1)
    Non_con_lake_cat = copy.copy(cat1)
    Non_con_lake_cat[:, :] = -9999
    arlakeid = np.unique(lake)
    arlakeid = arlakeid[arlakeid > 0]
    noncnlake = np.unique(noncnlake)
    noncnlake = noncnlake[noncnlake > 0]
    for i in range(0, len(arlakeid)):
        lakeid = arlakeid[i]
        lrowcol = np.argwhere(lake == lakeid).astype(int)
        lakacc = np.full((len(lrowcol), 3), -9999)
        lakacc[:, 0] = lrowcol[:, 0]
        lakacc[:, 1] = lrowcol[:, 1]
        lakacc[:, 2] = fac[lrowcol[:, 0], lrowcol[:, 1]]
        lakacc = lakacc[lakacc[:, 2].argsort()]
        lorow = lakacc[len(lakacc) - 1, 0]
        locol = lakacc[len(lakacc) - 1, 1]  ###### lake outlet row and col
        arclakeid = cat1[lorow, locol]
        pp = Pourpoints[lorow, locol]
        pp = np.unique(pp)
        pp = pp[pp > 0]
        #        print(lakeid,len(np.argwhere(noncnlake==lakeid)),pp,Pourpoints[lorow,locol],arclakeid)
        if len(pp) == 1:
            if arclakeid < 0:
                cat[lrowcol[:, 0], lrowcol[:, 1]] = pp
            else:
                cat[lrowcol[:, 0], lrowcol[:, 1]] = arclakeid
            if (
                len(np.argwhere(noncnlake == lakeid)) > 0
            ):  ##if lake belong to non connected lakes
                if arclakeid < 0:
                    nonlrowcol = np.argwhere(cat == pp).astype(int)
                    strids = np.unique(str_array[nonlrowcol[:, 0], nonlrowcol[:, 1]])
                    if (
                        len(strids) <= 1
                    ):  ## if none connected lake catchment overlay with stream, remove this lake
                        Non_con_lake_cat[nonlrowcol[:, 0], nonlrowcol[:, 1]] = lakeid
                    else:
                        cat[lrowcol[:, 0], lrowcol[:, 1]] = -99
                else:
                    nonlrowcol = np.argwhere(cat == arclakeid).astype(int)
                    strids = np.unique(str_array[nonlrowcol[:, 0], nonlrowcol[:, 1]])

                    if (
                        len(strids) <= 1
                    ):  ## if none connected lake catchment overlay with stream, remove this lake
                        Non_con_lake_cat[nonlrowcol[:, 0], nonlrowcol[:, 1]] = lakeid
                    else:
                        cat[lrowcol[:, 0], lrowcol[:, 1]] = -99
    #### For some reason pour point was missing in non-contribute catchment
    Pours = np.unique(Pourpoints)
    Pours = Pours[Pours > 0]
    for i in range(0, len(Pours)):
        pourid = Pours[i]
        rowcol = Pourpoints == pourid
        if cat[rowcol] < 0:
            temp_notused, nout = Is_Point_Close_To_Id_In_Raster(
                rowcol[0, 0], rowcol[0, 1], nrows, ncols, pourid, cat
            )
            if len(cat[cat == pourid]) > 0 and nout < 8:
                cat[rowcol] = pourid
    rowcol1 = fac > 0
    rowcol2 = cat < 0
    noncontribuite = np.logical_and(rowcol1, rowcol2)
    cat[noncontribuite] = 2 * max(np.unique(cat)) + 1
    return cat, Non_con_lake_cat


def GenerateFinalPourpoints(
    fac, fdir, lake, cat3, bsid, blid, boid, nrows, ncols, cat, obs, Is_divid_region
):
    Poups = copy.copy(cat3)
    Poups[:, :] = -9999
    GWat = copy.copy(cat3)
    GWatids = np.unique(cat3)
    GWatids = GWatids[GWatids >= 0]
    ncatid = 1
    for i in range(0, len(GWatids)):
        trow, tcol = Getbasinoutlet(GWatids[i], GWat, fac, fdir, nrows, ncols)
        #        arcpy.AddMessage("result     " +str(trow) +"    " +str(tcol))
        Poups[trow, tcol] = ncatid
        ncatid = ncatid + 1
    OWat = copy.copy(cat)
    OWatids = np.unique(cat)
    OWatids = OWatids[OWatids >= 0]
    for i in range(0, len(OWatids)):
        trow, tcol = Getbasinoutlet(OWatids[i], OWat, fac, fdir, nrows, ncols)
        if not GWat[trow, tcol] >= blid:
            if Poups[trow, tcol] < 0:
                Poups[trow, tcol] = ncatid
                ncatid = ncatid + 1
    if Is_divid_region > 0:
        return Poups
    obsids = np.unique(obs)
    obsids = obsids[obsids > 0]
    for i in range(0, len(obsids)):
        rowcol = np.argwhere(obs == obsids[i]).astype(int)
        if (
            Poups[rowcol[0, 0], rowcol[0, 1]] < 0
            and lake[rowcol[0, 0], rowcol[0, 1]] < 0
        ):
            Poups[rowcol[0, 0], rowcol[0, 1]] = ncatid
            ncatid = ncatid + 1
    return Poups


def Check_If_Str_Is_Head_Stream(prow, pcol, nrows, ncols, str, strid):

    noout = True
    ### the point prow and pcol is not surrounded by oter strems, except
    ### stream with stream id = strid
    if prow != 0 and prow != nrows - 1 and pcol != 0 and pcol != ncols - 1:

        if str[prow - 1, pcol + 1] > 0 and str[prow - 1, pcol + 1] != strid:
            noout = False
        if str[prow - 1, pcol - 1] > 0 and str[prow - 1, pcol - 1] != strid:
            noout = False
        if str[prow - 1, pcol] > 0 and str[prow - 1, pcol] != strid:
            noout = False
        if str[prow, pcol + 1] > 0 and str[prow, pcol + 1] != strid:
            noout = False
        if str[prow, pcol - 1] > 0 and str[prow, pcol - 1] != strid:
            noout = False
        if str[prow + 1, pcol - 1] > 0 and str[prow + 1, pcol - 1] != strid:
            noout = False
        if str[prow + 1, pcol + 1] > 0 and str[prow + 1, pcol + 1] != strid:
            noout = False
        if str[prow + 1, pcol] > 0 and str[prow + 1, pcol] != strid:
            noout = False
    return noout


###
def GenerPourpoint(
    cat,
    lake,
    Str,
    nrows,
    ncols,
    blid,
    bsid,
    bcid,
    fac,
    hydir,
    Is_divid_region,
    Remove_Str,
    Min_Grid_Number=50,
):
    GP_cat = copy.copy(cat)
    sblid = copy.copy(blid)
    Str_new = copy.copy(Str)
    ############### Part 1 Get all pourpoints of hydroshed catchment
    arcatid = np.unique(cat)  #### cat all catchment idd
    arcatid = arcatid[arcatid >= 0]
    catoutloc = np.full((len(arcatid), 3), -9999)
    for i in range(0, len(arcatid)):
        catid = arcatid[i]
        catrowcol = np.argwhere(cat == catid).astype(int)
        trow, tcol = Getbasinoutlet(catid, cat, fac, hydir, nrows, ncols)
        GP_cat[
            catrowcol[:, 0], catrowcol[:, 1]
        ] = -9999  ### set the catment cells into null
        GP_cat[trow, tcol] = bcid  #### change the outlet of catchment into wid
        bcid = bcid + 1
        catoutloc[i, 0] = catid  ## catchment id
        catoutloc[i, 1] = trow  #### catchment pourpont row
        catoutloc[i, 2] = tcol  #### catchment pourpont col
    #    writeraster(outFolder+subid+"_Pourpoints_1.asc",GP_cat)

    #### make the end point of stream in the head water shed to be a pourpoints

    Strids = np.unique(Str)
    Strids = Strids[Strids > 0]
    newstr_id = max(Strids) + 10
    for i in range(0, len(Strids)):
        strid = Strids[i]
        strrowcol = np.argwhere(Str == strid).astype(int)
        nstrrow = strrowcol.shape[0]

        ### if number of the stream grid smaller than 50

        if nstrrow < Min_Grid_Number:
            continue
        iStr = np.full((nstrrow, 4), -9999)  ##### 0 row, 1 col, 2 fac,
        iStr[:, 0] = strrowcol[:, 0]
        iStr[:, 1] = strrowcol[:, 1]
        iStr[:, 2] = fac[strrowcol[:, 0], strrowcol[:, 1]]
        iStr = iStr[iStr[:, 2].argsort()].astype(int)
        ### starting point of the stream
        Start_Pt_row = iStr[0, 0]
        Start_Pt_col = iStr[0, 1]

        Is_Head_Stream = Check_If_Str_Is_Head_Stream(
            Start_Pt_row, Start_Pt_col, nrows, ncols, Str, strid
        )
        if Is_Head_Stream == True:
            Str_new[
                iStr[0, 0], iStr[0, 1]
            ] = newstr_id  #### change the outlet of catchment into wid
            Str_new[iStr[1, 0], iStr[1, 1]] = newstr_id
            Str_new[iStr[2, 0], iStr[2, 1]] = newstr_id
            Str_new[iStr[3, 0], iStr[3, 1]] = newstr_id
            Str_new[iStr[4, 0], iStr[4, 1]] = newstr_id
            newstr_id = newstr_id + 1
    ######
    ##################Part 2 Get pourpoints of Lake inflow streams
    arlakeid = np.unique(lake)
    arlakeid = arlakeid[arlakeid >= 0]
    Lakemorestream = copy.copy(arlakeid)
    Lakemorestream[:] = -9999
    idxlakemorestream = 0
    for i in range(0, len(arlakeid)):  #### loop for each lake
        lid = arlakeid[i]  ##### obtain lake id
        rowcol = np.argwhere(lake == lid.astype(int))  ### got row anc col of lake grids
        nrow = rowcol.shape[0]  ### number of grids of the lake
        Stridinlake = np.full(nrow, -9999)  ####
        Stridinlake[:] = Str[
            rowcol[:, 0], rowcol[:, 1]
        ]  ### get all stream with in the lake domain
        Strid_L = np.unique(
            Stridinlake[np.argwhere(Stridinlake > 0).astype(int)]
        )  ### Get unique stream id of sach stream
        ##### find the intercept point of stream and lake
        if Is_divid_region > 0:
            if len(Strid_L) <= 1:
                Lakemorestream[idxlakemorestream] = lid
                idxlakemorestream = idxlakemorestream + 1
                continue

        for j in range(0, len(Strid_L)):  #### loop for each stream intercept with lake
            strid = Strid_L[j]
            ### get [row,col] of strid in str grids
            strrowcol = np.argwhere(Str == strid).astype(int)
            ### get number of grids of this str
            nstrrow = strrowcol.shape[0]
            ### create an array 0 will store row 1 will be col and 2 will be acc
            ### 3 will be the lake id
            Strchek = np.full((nstrrow, 4), -9999)  ##### 0 row, 1 col, 2 fac,
            Strchek[:, 0] = strrowcol[:, 0]
            Strchek[:, 1] = strrowcol[:, 1]
            Strchek[:, 2] = fac[strrowcol[:, 0], strrowcol[:, 1]]
            Strchek[:, 3] = lake[strrowcol[:, 0], strrowcol[:, 1]]
            ### sort the grids by flow accumulation
            Strchek = Strchek[Strchek[:, 2].argsort()].astype(int)

            ### find the first intersection point between river and lake
            Grids_Lake_river = np.where(Strchek[:, 3] == lid)
            irowst = np.nanmin(Grids_Lake_river)
            Intersection_Grid_row = Strchek[irowst, 0]
            Intersection_Grid_col = Strchek[irowst, 1]

            noout = 0
            ### double check if the head stream cell is nearby the lake
            if (
                Strchek[0, 0] != 0
                and Strchek[0, 0] != nrows - 1
                and Strchek[0, 1] != 0
                and Strchek[0, 1] != ncols - 1
            ):
                noout, temp_notused = Is_Point_Close_To_Id_In_Raster(
                    Strchek[0, 0], Strchek[0, 1], nrows, ncols, lid, lake
                )
            ####

            if irowst != 0:  ## is the stream is not start within the lake
                if (
                    lake[Strchek[irowst - 1, 0], Strchek[irowst - 1, 1]] == -9999
                ):  ### double check
                    ##### this means the stream connect two lakes, so must assign an pourpoints
                    if (
                        len(np.unique(lake[Strchek[0:irowst, 0], Strchek[0:irowst, 1]]))
                        >= 3
                    ):
                        GP_cat[Strchek[irowst - 1, 0], Strchek[irowst - 1, 1]] = bsid
                        bsid = bsid + 1

                    ##### the head stream celll is not nearby the lake
                    if noout == False:
                        GP_cat[Strchek[irowst - 1, 0], Strchek[irowst - 1, 1]] = bsid
                        bsid = bsid + 1
            #### it is possible that two steam combine together near the lake, double check if the stream conncet to
            # anotehr stream and this steam is not witin the lake
            if irowst == 0 or noout == True:
                nostr = Str[Strchek[0, 0], Strchek[0, 1]]
                a = 0
                orowcol = np.full((8, 3), -9999)
                if (
                    Strchek[0, 0] != 0
                    and Strchek[0, 0] != nrows - 1
                    and Strchek[0, 1] != 0
                    and Strchek[0, 1] != ncols - 1
                ):
                    if (
                        Str[Strchek[0, 0] - 1, Strchek[0, 1] + 1] != -9999
                        and lake[Strchek[0, 0] - 1, Strchek[0, 1] + 1] == -9999
                    ):
                        orowcol[a, 0] = Strchek[0, 0] - 1
                        orowcol[a, 1] = Strchek[0, 1] + 1
                        orowcol[a, 2] = Str[Strchek[0, 0] - 1, Strchek[0, 1] + 1]
                        a = a + 1
                    if (
                        Str[Strchek[0, 0] - 1, Strchek[0, 1] - 1] != -9999
                        and lake[Strchek[0, 0] - 1, Strchek[0, 1] - 1] == -9999
                    ):
                        orowcol[a, 0] = Strchek[0, 0] - 1
                        orowcol[a, 1] = Strchek[0, 1] - 1
                        orowcol[a, 2] = Str[Strchek[0, 0] - 1, Strchek[0, 1] - 1]
                        a = a + 1
                    if (
                        Str[Strchek[0, 0] - 1, Strchek[0, 1]] != -9999
                        and lake[Strchek[0, 0] - 1, Strchek[0, 1]] == -9999
                    ):
                        orowcol[a, 0] = Strchek[0, 0] - 1
                        orowcol[a, 1] = Strchek[0, 1] - 0
                        orowcol[a, 2] = Str[Strchek[0, 0] - 1, Strchek[0, 1] - 0]
                        a = a + 1
                    if (
                        Str[Strchek[0, 0], Strchek[0, 1] + 1] != -9999
                        and lake[Strchek[0, 0], Strchek[0, 1] + 1] == -9999
                    ):
                        orowcol[a, 0] = Strchek[0, 0] - 0
                        orowcol[a, 1] = Strchek[0, 1] + 1
                        orowcol[a, 2] = Str[Strchek[0, 0] - 0, Strchek[0, 1] + 1]
                        a = a + 1
                    if (
                        Str[Strchek[0, 0], Strchek[0, 1] - 1] != -9999
                        and lake[Strchek[0, 0], Strchek[0, 1] - 1] == -9999
                    ):
                        orowcol[a, 0] = Strchek[0, 0] - 0
                        orowcol[a, 1] = Strchek[0, 1] - 1
                        orowcol[a, 2] = Str[Strchek[0, 0] - 0, Strchek[0, 1] - 1]
                        a = a + 1
                    if (
                        Str[Strchek[0, 0] + 1, Strchek[0, 1] - 1] != -9999
                        and lake[Strchek[0, 0] + 1, Strchek[0, 1] - 1] == -9999
                    ):
                        orowcol[a, 0] = Strchek[0, 0] + 1
                        orowcol[a, 1] = Strchek[0, 1] - 1
                        orowcol[a, 2] = Str[Strchek[0, 0] + 1, Strchek[0, 1] - 1]
                        a = a + 1
                    if (
                        Str[Strchek[0, 0] + 1, Strchek[0, 1] + 1] != -9999
                        and lake[Strchek[0, 0] + 1, Strchek[0, 1] + 1] == -9999
                    ):
                        orowcol[a, 0] = Strchek[0, 0] + 1
                        orowcol[a, 1] = Strchek[0, 1] + 1
                        orowcol[a, 2] = Str[Strchek[0, 0] + 1, Strchek[0, 1] + 1]
                        a = a + 1
                    if (
                        Str[Strchek[0, 0] + 1, Strchek[0, 1]] != -9999
                        and lake[Strchek[0, 0] + 1, Strchek[0, 1]] == -9999
                    ):
                        orowcol[a, 0] = Strchek[0, 0] + 1
                        orowcol[a, 1] = Strchek[0, 1] - 0
                        orowcol[a, 2] = Str[Strchek[0, 0] + 1, Strchek[0, 1] - 0]
                        a = a + 1
                if a > 0:
                    for ka in range(0, a):
                        nostr = orowcol[ka, 2]  ## up stream stram id
                        srowcol = np.argwhere(Str == nostr).astype(int)
                        snrow = srowcol.shape[0]
                        iStrchek = np.full(
                            (snrow, 4), -9999
                        )  ##### 0 row, 1 col, 2 fac,
                        iStrchek[:, 0] = srowcol[:, 0]
                        iStrchek[:, 1] = srowcol[:, 1]
                        iStrchek[:, 2] = fac[srowcol[:, 0], srowcol[:, 1]]
                        iStrchek = iStrchek[iStrchek[:, 2].argsort()]
                        noout, temp_notused = Is_Point_Close_To_Id_In_Raster(
                            iStrchek[0, 0], iStrchek[0, 1], nrows, ncols, lid, lake
                        )
                        Lakinstr = np.full(snrow, -9999)
                        Lakinstr[:] = lake[srowcol[:, 0], srowcol[:, 1]]
                        d = np.argwhere(Lakinstr == lid).astype(
                            int
                        )  #### the connected stream should not within the lake
                        if len(d) < 1 and noout == False:
                            GP_cat[orowcol[ka, 0], orowcol[ka, 1]] = bsid
                            bsid = bsid + 1
    ################################################################################
    ################## Part 3Get Lake pourpoint id and remove cat pourpoint that contribute to lake
    #    writeraster(outFolder+"Pourpoints_2.asc",GP_cat)
    Lakemorestream = Lakemorestream[Lakemorestream > 0]
    for i in range(0, len(arlakeid)):
        lakeid = arlakeid[i]

        if Is_divid_region > 0:
            mask = Lakemorestream == lakeid
            if np.sum(mask) >= 1:
                continue

        lrowcol = np.argwhere(lake == lakeid).astype(int)
        arcatid = np.unique(
            cat[lrowcol[:, 0], lrowcol[:, 1]]
        )  ## Get all catchment that intercept with lake
        lakeacc = np.full((len(lrowcol), 3), -9999)
        lakeacc[:, 0] = lrowcol[:, 0]
        lakeacc[:, 1] = lrowcol[:, 1]
        lakeacc[:, 2] = fac[lrowcol[:, 0], lrowcol[:, 1]]
        lakeacc = lakeacc[lakeacc[:, 2].argsort()]
        maxcatid = cat[
            lakeacc[len(lakeacc) - 1, 0], lakeacc[len(lakeacc) - 1, 1]
        ]  ### Get the catchment id that will keeped

        ### remove the all pourpoints close to lake pourpoints
        if (
            lakeacc[len(lakeacc) - 1, 0] != 0
            and lakeacc[len(lakeacc) - 1, 0] != nrows - 1
            and lakeacc[len(lakeacc) - 1, 1] != 0
            and lakeacc[len(lakeacc) - 1, 1] != ncols - 1
        ):
            GP_cat[
                lakeacc[len(lakeacc) - 1, 0] - 1, lakeacc[len(lakeacc) - 1, 1] - 1
            ] = -9999
            GP_cat[
                lakeacc[len(lakeacc) - 1, 0] + 1, lakeacc[len(lakeacc) - 1, 1] - 1
            ] = -9999
            GP_cat[
                lakeacc[len(lakeacc) - 1, 0], lakeacc[len(lakeacc) - 1, 1] - 1
            ] = -9999
            GP_cat[
                lakeacc[len(lakeacc) - 1, 0] + 1, lakeacc[len(lakeacc) - 1, 1]
            ] = -9999
            GP_cat[
                lakeacc[len(lakeacc) - 1, 0] + 1, lakeacc[len(lakeacc) - 1, 1] + 1
            ] = -9999
            GP_cat[
                lakeacc[len(lakeacc) - 1, 0], lakeacc[len(lakeacc) - 1, 1] + 1
            ] = -9999
            GP_cat[
                lakeacc[len(lakeacc) - 1, 0] - 1, lakeacc[len(lakeacc) - 1, 1]
            ] = -9999

        ### remove pourpoints that not equal to maxcatid
        for j in range(0, len(arcatid)):
            if arcatid[j] != maxcatid:
                crowcol = np.argwhere(cat == arcatid[j]).astype(int)
                checkcat = np.full((len(crowcol), 4), -9999)
                checkcat[:, 0] = crowcol[:, 0]
                checkcat[:, 1] = crowcol[:, 1]
                checkcat[:, 2] = GP_cat[crowcol[:, 0], crowcol[:, 1]]
                checkcat[:, 3] = fac[crowcol[:, 0], crowcol[:, 1]]
                checkcat = checkcat[checkcat[:, 3].argsort()]
                dele = checkcat[checkcat[:, 2] < blid].astype(
                    int
                )  #### do not delete the lake and strem ids
                GP_cat[dele[:, 0], dele[:, 1]] = -9999

                ### add keep catchment in case the catchment outlet is at the boundary of the domain
                nrow, ncol = Nextcell(
                    hydir,
                    checkcat[len(checkcat) - 1, 0],
                    checkcat[len(checkcat) - 1, 1],
                )
                if nrow > 0 or ncol > 0:
                    if nrow >= nrows or ncols >= ncols:
                        continue
                    if cat[nrow, ncol] < 0:
                        GP_cat[
                            checkcat[len(checkcat) - 1, 0],
                            checkcat[len(checkcat) - 1, 1],
                        ] = bcid
                        bcid = bcid + 1

        #### add lake outlet
        GP_cat[lakeacc[len(lakeacc) - 1, 0], lakeacc[len(lakeacc) - 1, 1]] = sblid
        sblid = sblid + 1
    return GP_cat, Lakemorestream, Str_new


def Getbasinoutlet(ID, basin, fac, dir, nrows, ncols):
    catrowcol = np.argwhere(basin == ID).astype(int)
    catacc = np.full((len(catrowcol), 3), -9999)
    catacc[:, 0] = catrowcol[:, 0]
    catacc[:, 1] = catrowcol[:, 1]
    catacc[:, 2] = fac[catrowcol[:, 0], catrowcol[:, 1]]
    catacc = catacc[catacc[:, 2].argsort()]
    #    print(ID,len(catrowcol))
    ### check if it is a real basin outlet
    crow = catacc[len(catrowcol) - 1, 0]
    ccol = catacc[len(catrowcol) - 1, 1]

    nrow, ncol = Nextcell(dir, crow, ccol)
    #    print(ID,basin[nrow,ncol],basin[crow,ccol],fac[nrow,ncol],fac[crow,ccol],crow,ccol)
    if nrow < 0 or ncol < 0:
        return crow, ccol
    elif nrow >= nrows or ncol >= ncols:
        return crow, ccol
    elif basin[nrow, ncol] < 0:
        return crow, ccol
    elif basin[nrow, ncol] != ID:  #  all above means the outlet is the real loutlet
        return crow, ccol
    else:
        crow = nrow
        ccol = ncol
        ifound = 0
        for i in range(0, 1000):  #### find next 1000 grids, to find the basin outlet
            nrow, ncol = Nextcell(dir, crow, ccol)
            if nrow < 0 or ncol < 0:
                ifound = 1
                break
            elif nrow >= nrows or ncol >= ncols:
                ifound = 1
                break
            elif basin[nrow, ncol] < 0:
                ifound = 1
                break
            elif basin[nrow, ncol] != ID:
                ifound = 1  #     all above means the outlet is the real loutlet
                break
            else:
                crow = nrow
                ccol = ncol
                continue
        if ifound == 0:
            print(" true basin outlet not found for ID...." + str(ID))
        return crow, ccol


def Nextcell(N_dir, N_row, N_col):
    if N_dir[N_row, N_col] == 1:
        N_nrow = N_row + 0
        N_ncol = N_col + 1
    elif N_dir[N_row, N_col] == 2:
        N_nrow = N_row + 1
        N_ncol = N_col + 1
    elif N_dir[N_row, N_col] == 4:
        N_nrow = N_row + 1
        N_ncol = N_col + 0
    elif N_dir[N_row, N_col] == 8:
        N_nrow = N_row + 1
        N_ncol = N_col - 1
    elif N_dir[N_row, N_col] == 16:
        N_nrow = N_row + 0
        N_ncol = N_col - 1
    elif N_dir[N_row, N_col] == 32:
        N_nrow = N_row - 1
        N_ncol = N_col - 1
    elif N_dir[N_row, N_col] == 64:
        N_nrow = N_row - 1
        N_ncol = N_col + 0
    elif N_dir[N_row, N_col] == 128:
        N_nrow = N_row - 1
        N_ncol = N_col + 1
    else:
        N_nrow = -9999
        N_ncol = -9999
    return N_nrow, N_ncol


def return_subid_of_next_down_stream_grids(
    cat_array, catid, nfdr_arcgis_array, cat_outlet_array, ncols, nrows
):
    # get outlet row cols
    maxtry = 10
    outlet_row_col = np.argwhere(cat_outlet_array == catid)
    if len(outlet_row_col) <= 0:
        return -1
    outlet_row = outlet_row_col[0, 0]
    outlet_col = outlet_row_col[0, 1]
    downcatid = -1
    for i in range(0, maxtry):
        nrow, ncol = Nextcell(nfdr_arcgis_array, outlet_row, outlet_col)
        if nrow >= nrows or ncol >= ncols:
            downsubid = -1
            return downsubid
        elif nrow <= 0 or ncol <= 0:
            downsubid = -1
            return downsubid
        else:
            downcatid = cat_array[nrow, ncol]
            if downcatid != catid:
                return downcatid
            else:
                outlet_row = nrow
                outlet_col = ncol
    return downcatid



from basinmaker.func.arcgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def add_lakes_into_existing_watershed_delineation(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    path_lakefile_in,
    lake_attributes,
    threshold_con_lake,
    threshold_non_con_lake,
    only_included_lake_at_river_interction = False,
    remove_lake_inlets = False,
    path_sub_reg_lake_r="#",
    path_sub_reg_lake_bd_r="#",
    sl_connected_lake="sl_connected_lake",
    sl_non_connected_lake="sl_nonconnect_lake",
    sl_lakes="selected_lakes",
    sl_str_connected_lake="str_sl_connected_lake",
    nfdr_arcgis="narcgis_fdr",
    nfdr_grass="ngrass_fdr",
    cat_add_lake="cat_add_lake",
    pourpoints_with_lakes="pourpoints_with_lakes",
    cat_use_default_acc="cat_use_default_acc",
    lake_outflow_pourpoints="lake_outflow_pourpoints",
    problem_seg="problem_seg",
    max_memroy=1024 * 4,
):
    work_folder = grassdb
    # define required input files names
    fdr_arcgis = input_geo_names["fdr_arcgis"]
    fdr_grass = input_geo_names["fdr_grass"]
    str_r = input_geo_names["str_r"]
    str_v = input_geo_names["str_v"]
    acc = input_geo_names["acc"]
    cat_no_lake = input_geo_names["cat_no_lake"]
    mask = input_geo_names["mask"]
    dem = input_geo_names["dem"]

    # define internal file names
    lake_inflow_pourpoints = Internal_Constant_Names["lake_inflow_pourpoints"]
    catchment_pourpoints_outside_lake = Internal_Constant_Names[
        "catchment_pourpoints_outside_lake"
    ]
    cat_add_lake_old_fdr = Internal_Constant_Names["cat_add_lake_old_fdr"]
    str_connected_lake = Internal_Constant_Names["str_connected_lake"]
    alllake = Internal_Constant_Names["all_lakes"]
    lake_boundary = Internal_Constant_Names["lake_boundary"]
    connected_lake = Internal_Constant_Names["connect_lake"]
    non_connected_lake = Internal_Constant_Names["nonconnect_lake"]
    lakes_lg_cl_thres = "lakes_lg_cl_thres"
    lakes_lg_ncl_thres = "lakes_lg_ncl_thres"


    if not os.path.exists(work_folder):
        os.makedirs(work_folder)
    arcpy.env.workspace = work_folder
     
    arcpy.env.overwriteOutput = True
    arcpy.CheckOutExtension("Spatial")
    cellSize = float(arcpy.GetRasterProperties_management(dem+'.tif', "CELLSIZEX").getOutput(0))
    SptailRef = arcpy.Describe(dem+'.tif').spatialReference
    arcpy.env.XYTolerance = cellSize
    arcpy.arcpy.env.cellSize = cellSize    
    arcpy.env.outputCoordinateSystem = arcpy.SpatialReference(int(SptailRef.factoryCode)) ### WGS84
    arcpy.env.extent = arcpy.Describe(dem + '.tif').extent
    
    ## processing lakes move to a function later 
    
    ## obtain lakes fully contained in the mask region 
    ## any lake overlay with the mask boundary will be removed 
    arcpy.Project_management(
        path_lakefile_in,
        "lake_proj.shp", 
        arcpy.SpatialReference(int(SptailRef.factoryCode)),
        )  
        
    arcpy.FeatureToLine_management(mask + '.shp', mask + '_line.shp')
    arcpy.Clip_analysis("lake_proj.shp", mask + '.shp', "lake_clip.shp", "")
    arcpy.Intersect_analysis(["lake_proj.shp",mask + '_line.shp'], 'lake_inter.shp')
    inter_lake = pd.DataFrame.spatial.from_featureclass(os.path.join(work_folder,'lake_inter.shp'))
    all_cliped_lakes = pd.DataFrame.spatial.from_featureclass(os.path.join(work_folder,'lake_clip.shp'))
    lakeids_inter = inter_lake[lake_attributes[0]].values
    select_lake = all_cliped_lakes.loc[~all_cliped_lakes[lake_attributes[0]].isin(lakeids_inter)]
    select_lake.spatial.to_featureclass(location=os.path.join(work_folder,alllake + ".shp"),overwrite=True,sanitize_columns=False)    
    ###
    arcpy.FeatureToLine_management(alllake + '.shp', alllake + '_line.shp')
    
    arcpy.Dissolve_management(alllake + '_line.shp', lake_boundary + ".shp", [lake_attributes[0]])
    
    arcpy.PolygonToRaster_conversion(alllake + ".shp", lake_attributes[0], alllake + ".tif",
                                 "MAXIMUM_COMBINED_AREA",lake_attributes[0], cellSize)
                                 
    arcpy.PolylineToRaster_conversion(lake_boundary + ".shp", lake_attributes[0], lake_boundary + ".tif",
                                 "MAXIMUM_COMBINED_LENGTH",lake_attributes[0], cellSize)
                                 
    return



import copy
import os

import numpy as np


def modify_lakes_flow_direction(
    cat3,
    lake,
    fac,
    dir,
    str_array,
    lakeinfo,
    nrows,
    ncols,
    LakeBD_array,
    Pec_Grid_outlier,
    MaximumLakegrids,
    Lakemorestream,
):
    cat = copy.copy(cat3)
    ndir = copy.copy(dir)
    changed_ndir = copy.copy(dir)
    changed_ndir[:, :] = -9999
    BD_problem = copy.copy(dir)
    BD_problem[:, :] = -9999
    arlakeid = np.unique(lake)
    arlakeid = arlakeid[arlakeid > 0]
    outlakeids = np.full((1000000, 2), -99999.999)
    stream_mask = str_array > 0
    # sort with lake outlet flow accumulation thresthold
    lakeinfo = lakeinfo.loc[lakeinfo['cat'].isin(arlakeid)]
    lakeinfo = lakeinfo.sort_values(by='lmax_acc', ascending=False)

    for i in range(0, len(lakeinfo)):
        lakeid = lakeinfo['cat'].values[i]
        if lakeid in Lakemorestream:
            continue

        lrowcol = np.argwhere(lake == lakeid).astype(int)
        
        # get minimum and maximum row and col number 
        lakerows = np.arange(max(np.amin(lrowcol[:, 0])-10,0),min(np.amax(lrowcol[:, 0])+10,nrows-1))
        lakecols = np.arange(max(np.amin(lrowcol[:, 1])-10,0),min(np.amax(lrowcol[:, 1])+10,ncols-1))

        # create a index grid based on col and rows of lake id
        ixgrid = np.ix_(lakerows, lakecols)
        
        # extract sub np array 
        lake_sub_lake = lake[ixgrid] 
        lake_sub_fac = fac[ixgrid] 
        lake_sub_cat = cat[ixgrid] 
        lake_sub_changed_ndir = changed_ndir[ixgrid] 
        lake_sub_BD_problem = BD_problem[ixgrid] 
        lake_sub_ndir = ndir[ixgrid] 
        lake_sub_LakeBD = LakeBD_array[ixgrid]
        
        # obtain changed flow direction 
        lake_sub_ndir,lake_sub_BD_problem,lake_sub_changed_ndir=change_lake_flow_direction(
            lakeid,
            lake_sub_lake,
            lake_sub_fac,
            lake_sub_cat,
            lake_sub_changed_ndir,
            lake_sub_BD_problem,
            lake_sub_ndir,
            lake_sub_LakeBD,
            Pec_Grid_outlier,
            lake_sub_lake.shape[0],
            lake_sub_lake.shape[1], 
            lakeinfo,  
            MaximumLakegrids, 
            i,    
        )
        
        # copy changed flow direction to original array 
        ndir[ixgrid] = lake_sub_ndir
        changed_ndir[ixgrid] = lake_sub_changed_ndir
        BD_problem[ixgrid] = lake_sub_BD_problem
        
    return 'aaa',changed_ndir, ndir, BD_problem

def change_lake_flow_direction(
    lakeid,
    lake_sub_lake,
    lake_sub_fac,
    lake_sub_cat,
    lake_sub_changed_ndir,
    lake_sub_BD_problem,
    lake_sub_ndir,
    lake_sub_LakeBD,
    Pec_Grid_outlier,
    nrows,
    ncols,
    lakeinfo,
    MaximumLakegrids,
    i,
):
    lrowcol = np.argwhere(lake_sub_lake == lakeid).astype(int)
    lakacc = np.full((len(lrowcol), 3), -9999)
    lakacc[:, 0] = lrowcol[:, 0]
    lakacc[:, 1] = lrowcol[:, 1]
    lakacc[:, 2] = lake_sub_fac[lrowcol[:, 0], lrowcol[:, 1]]
    lakacc = lakacc[lakacc[:, 2].argsort()]
    ###### lake outlet row and col
    lorow = lakacc[len(lakacc) - 1, 0]
    locol = lakacc[len(lakacc) - 1, 1]  
    ####### lake catchment id
    arclakeid = lake_sub_cat[lorow, locol] 
    # catchment row and col  
    lakecatrowcol = np.argwhere(lake_sub_cat == arclakeid).astype(int) 
     
    #### lake catchment cells
    Lakeincat1 = lake_sub_lake[lakecatrowcol[:, 0], lakecatrowcol[:, 1]]
    nlake = np.argwhere(Lakeincat1 == lakeid).astype(int)
    
    #        print("########################################################################3")
    #        print(lakeid,arclakeid,len(nlake),len(lrowcol),float(len(nlake)/len(lrowcol)),Pec_Grid_outlier)
    if float(len(nlake) / len(lrowcol))  > Pec_Grid_outlier:
        return lake_sub_ndir,lake_sub_BD_problem,lake_sub_changed_ndir

    #        if outlakeids[i,1] > Pec_Grid_outlier: ### smaller than 0.97
    #            continue

    #        print(lakeid,arclakeid,len(nlake),len(lrowcol),float(len(nlake)/len(lrowcol)))
    
    BD_mask = lake_sub_LakeBD == lakeid
    Lake_mask = lake_sub_lake == lakeid
    cat_lake_mask = lake_sub_cat == arclakeid
    
    # mask grids that belong to both lake and lake cat 
    Lakeincat_mask = np.logical_and(Lake_mask, cat_lake_mask)
    nlake2 = np.sum(Lakeincat_mask)
     
    # mask grids that do not in the lake cat 
    Lakeoutcat_mask = np.logical_and(Lake_mask, np.logical_not(cat_lake_mask))
    
    # mask lake boundary grids that not in the lake cat 
    BD_Out_Lakecat_mask = np.logical_and(Lakeoutcat_mask, BD_mask)

    #        BD_Out_Lakecat_Nriv_mask = np.logical_and(BD_Out_Lakecat_mask,np.logical_not(stream_mask)) do not allow modify stream grids
    BD_Out_Lakecat_Nriv_mask = BD_Out_Lakecat_mask
    lake_sub_BD_problem[BD_Out_Lakecat_Nriv_mask] = 1
    
     ### smaller than nlakegrids or smaller than 0.9
    if (np.sum(BD_Out_Lakecat_mask) > MaximumLakegrids): 
        return lake_sub_ndir,lake_sub_BD_problem,lake_sub_changed_ndir

    print("Total # of lakes: ", len(lakeinfo), " Processing    ", i, "th lake")
    print(
        "Lake ID : ",
        lakeid,
        "Lake Cat ID   ",
        arclakeid,
        "Total numer of Lake grids   ",
        len(lrowcol),
        "Numer of Lake grids in Lake Cat:  ",
        nlake2,
        len(nlake),
    )
    print(
        "# of Lake boundary grids:   ",
        np.sum(BD_mask),
        "# of grids do not flow to lake catchments",
        np.sum(Lakeoutcat_mask),
        "# of lake boundary grids not flow to lake catchment   ",
        np.sum(BD_Out_Lakecat_mask),
        "  # of lake boundary grids not flow to lake catchment not a river gird  ",
        np.sum(BD_Out_Lakecat_Nriv_mask),
    )

    #####  Locate the grids that at the target grids ege
    Grid_Nee_Mo = np.argwhere(BD_Out_Lakecat_Nriv_mask == 1)
    #        print(print(Grid_Nee_Mo[:,0], Grid_Nee_Mo[:,1]))

    # problem_points is the point needs to modfiy flow direction
    problem_points = copy.copy(Grid_Nee_Mo)
    idx_good = 0
    idx_problem = 0
    new_problem_points = np.full((len(problem_points) + 40, 2), -9999)
    good_starts_points = np.full((len(problem_points) + 40, 2), -9999)

    # first loop to find boundary points that can flow to lake catchment

    for i in range(0, len(problem_points)):
        # get problem point row anc col
        trow = problem_points[i, 0]
        tcol = problem_points[i, 1]
        #                print(i,ipo,trow,tcol)
        if trow >= nrows - 1 or tcol == ncols - 1:
            continue
        # try to modify the flow direction of this row and col
        # see if it can flow to the lake catchment

        lake_sub_ndir, IS_Change, lake_sub_changed_ndir = changeflowdirectionofedgegrids(
            lake_sub_ndir,
            trow,
            tcol,
            Lakeincat_mask,
            1,
            ncols,
            nrows,
            BD_Out_Lakecat_Nriv_mask,
            lake_sub_changed_ndir,
        )

        # if the point can flow to the lake catchment
        # store the points in to goodpoint2
        if IS_Change > 0:
            Lakeincat_mask[trow, tcol] = 1
            good_starts_points[idx_good, 0] = trow
            good_starts_points[idx_good, 1] = tcol
            idx_good = idx_good + 1
        else:
            new_problem_points[idx_problem, 0] = trow
            new_problem_points[idx_problem, 1] = tcol
            idx_problem = idx_problem + 1

    problem_points = copy.copy(new_problem_points)
    all_good_start_points = copy.copy(good_starts_points)
    idx_good_all = idx_good + 1
    iter = 0
    good_starts_points = good_starts_points[good_starts_points[:, 0] > 0]
    problem_points = problem_points[problem_points[:, 0] > 0]
    # print("#################################")
    # loop until all len problem_points is zero or iteration times exceed
    nproblempoint = len(problem_points)
    while (
        iter < nproblempoint + 10
        and len(problem_points) > 0
        and len(good_starts_points) > 0
    ):
        iter = iter + 1
        idx_good_loop = 0
        new_good_starts_points = np.full((len(BD_Out_Lakecat_mask) + 30, 2), -9999)
        # loop for each good points and modify flow diretion of grids around it
        # make them flow to good points
        # print(iter)
        # print(len(all_good_start_points[all_good_start_points[:,0] > 0]))
        # print(len(good_starts_points))
        # print(good_starts_points)

        for i in range(0, len(good_starts_points)):
            trow = good_starts_points[i, 0]
            tcol = good_starts_points[i, 1]
            # print(i,trow,tcol)
            # print(len(all_good_start_points[all_good_start_points[:,0] > 0]))
            # print(idx_good_all)
            # print(len(good_starts_points))
            # print(idx_good_loop)
            # print(len(problem_points))
            # print("a")

            if trow >= nrows - 1 or tcol == ncols - 1:
                continue
            (
                lake_sub_ndir,
                lake_sub_changed_ndir,
                new_good_starts_points,
                idx_good_loop,
                all_good_start_points,
                idx_good_all,
                problem_points,
                Lakeincat_mask,
            ) = modify_flow_direction_of_nearby_boundary_grids(
                lake_sub_ndir,
                trow,
                tcol,
                Lakeincat_mask,
                1,
                new_good_starts_points,
                all_good_start_points,
                idx_good_loop,
                idx_good_all,
                problem_points,
                BD_Out_Lakecat_Nriv_mask,
                lake_sub_changed_ndir,
            )

            # print("v")
            # print(new_good_starts_points[new_good_starts_points[:,0]>0])
            # print("###")
        # keep only larger
        good_starts_points = new_good_starts_points[
            new_good_starts_points[:, 0] > 0
        ]

    ## in case some point is missed, loop problem point again
    iter = 0
    nproblempoint = len(problem_points)
    while iter < nproblempoint + 10 and len(problem_points) > 0:

        # loop for each problem points, and check if it can flow to an lake
        # catchment, by chaning it's flow direction only.
        #
        iter = iter + 1
        idx = 0
        new_problem_points = np.full((len(problem_points) + 10, 2), -9999)
        for i in range(0, len(problem_points)):
            # get problem point row anc col
            trow = problem_points[i, 0]
            tcol = problem_points[i, 1]
            #                print(i,ipo,trow,tcol)
            if trow >= nrows - 1 or tcol == ncols - 1:
                continue

            # try to modify the flow direction of this row and col
            # see if it can flow to the lake catchment

            lake_sub_ndir, IS_Change, lake_sub_changed_ndir = changeflowdirectionofedgegrids(
                lake_sub_ndir,
                trow,
                tcol,
                Lakeincat_mask,
                1,
                ncols,
                nrows,
                BD_Out_Lakecat_Nriv_mask,
                lake_sub_changed_ndir,
            )

            # if the point can flow to the lake catchment
            # store the points in to goodpoint2
            if IS_Change > 0:
                Lakeincat_mask[trow, tcol] = 1
            else:
                new_problem_points[idx, 0] = trow
                new_problem_points[idx, 1] = tcol
                idx = idx + 1
        # keep only larger
        new_problem_points = new_problem_points[new_problem_points[:, 0] > 0]
        problem_points = new_problem_points

    print(" # of uncorrect lake boundary grids     ", len(problem_points))
    if len(problem_points) > 0:
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        print(lakeid)
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
    print("#####################################")
    return lake_sub_ndir,lake_sub_BD_problem,lake_sub_changed_ndir

def changeflowdirectionofedgegrids(
    N_dir,
    p_row,
    p_col,
    lake1,
    lid,
    ncols,
    nrows,
    BD_Out_Lakecat_Nriv_mask,
    Changed_ndir,
):
    ndir = copy.copy(N_dir)
    changed_ndir = copy.copy(Changed_ndir)
    IS_Change = 0

    if (
        lake1[p_row + 0, p_col + 1] == lid
    ):  #### it is a broudary grids and did not flow to the lake outlet
        ndir[p_row, p_col] = 1  ### change the flow direction of new point to old points
        changed_ndir[p_row, p_col] = 1
        IS_Change = 1
    # dir 2
    if lake1[p_row + 1, p_col + 1] == lid:

        ndir[p_row, p_col] = 2
        changed_ndir[p_row, p_col] = 2
        IS_Change = 1
    # dir 3
    if lake1[p_row + 1, p_col + 0] == lid:

        ndir[p_row, p_col] = 4
        changed_ndir[p_row, p_col] = 4
        IS_Change = 1
    # dir 4
    if lake1[p_row + 1, p_col - 1] == lid:

        ndir[p_row, p_col] = 8
        changed_ndir[p_row, p_col] = 8
        IS_Change = 1

    # dir 5
    if lake1[p_row + 0, p_col - 1] == lid:

        ndir[p_row, p_col] = 16
        changed_ndir[p_row, p_col] = 16
        IS_Change = 1

    # dir 6
    if lake1[p_row - 1, p_col - 1] == lid:

        ndir[p_row, p_col] = 32
        changed_ndir[p_row, p_col] = 32
        IS_Change = 1
    # dir 7
    if lake1[p_row - 1, p_col + 0] == lid:

        ndir[p_row, p_col] = 64
        changed_ndir[p_row, p_col] = 64
        IS_Change = 1
    # dir 8
    if lake1[p_row - 1, p_col + 1] == lid:

        ndir[p_row, p_col] = 128
        changed_ndir[p_row, p_col] = 128
        IS_Change = 1
    #    arcpy.AddMessage(goodpoint[goodpoint[:,0]>0,])
    return ndir, IS_Change, changed_ndir


def Dirpoints_v3(
    N_dir,
    p_row,
    p_col,
    lake1,
    lid,
    goodpoint,
    k,
    ncols,
    nrows,
    problem_points,
    Changed_ndir,
):
    ### this function change flow direction of some grids around p_row, p_col, flow to p_row, p_col

    ndir = copy.copy(N_dir)
    changed_ndir = copy.copy(Changed_ndir)
    ip = copy.copy(k) + 1
    # dir 1
    #### it is a broudary grids and did not flow to the lake outlet
    if BD_Out_Lakecat_Nriv_mask[p_row + 0, p_col + 1] == lid:
        tt = goodpoint[
            goodpoint[:, 0] == p_row + 0,
        ]
        #### the point not exist in good points which store all boundary points
        if (
            len(
                tt[
                    tt[:, 1] == p_col + 1,
                ]
            )
            < 1
        ):
            ### change the flow direction of new point to old points
            ndir[p_row + 0, p_col + 1] = 16
            changed_ndir[p_row + 0, p_col + 1] = 16
            goodpoint[ip, 0] = p_row + 0
            goodpoint[ip, 1] = p_col + 1
            ip = ip + 1
    # dir 2
    if BD_Out_Lakecat_Nriv_mask[p_row + 1, p_col + 1] == lid:

        tt = goodpoint[
            goodpoint[:, 0] == p_row + 1,
        ]
        if (
            len(
                tt[
                    tt[:, 1] == p_col + 1,
                ]
            )
            < 1
        ):
            ndir[p_row + 1, p_col + 1] = 32
            changed_ndir[p_row + 1, p_col + 1] = 32
            goodpoint[ip, 0] = p_row + 1
            goodpoint[ip, 1] = p_col + 1
            ip = ip + 1
    # dir 3
    if BD_Out_Lakecat_Nriv_mask[p_row + 1, p_col + 0] == lid:

        tt = goodpoint[
            goodpoint[:, 0] == p_row + 1,
        ]
        if (
            len(
                tt[
                    tt[:, 1] == p_col + 0,
                ]
            )
            < 1
        ):
            ndir[p_row + 1, p_col + 0] = 64
            changed_ndir[p_row + 1, p_col + 0] = 64
            goodpoint[ip, 0] = p_row + 1
            goodpoint[ip, 1] = p_col + 0
            ip = ip + 1
    # dir 4
    if BD_Out_Lakecat_Nriv_mask[p_row + 1, p_col - 1] == lid:

        tt = goodpoint[
            goodpoint[:, 0] == p_row + 1,
        ]
        if (
            len(
                tt[
                    tt[:, 1] == p_col - 1,
                ]
            )
            < 1
        ):
            ndir[p_row + 1, p_col - 1] = 128
            changed_ndir[p_row + 1, p_col - 1] = 128
            goodpoint[ip, 0] = p_row + 1
            goodpoint[ip, 1] = p_col - 1
            ip = ip + 1
    # dir 5
    if BD_Out_Lakecat_Nriv_mask[p_row + 0, p_col - 1] == lid:

        tt = goodpoint[
            goodpoint[:, 0] == p_row + 0,
        ]
        if (
            len(
                tt[
                    tt[:, 1] == p_col - 1,
                ]
            )
            < 1
        ):
            ndir[p_row + 0, p_col - 1] = 1
            changed_ndir[p_row + 0, p_col - 1] = 1
            goodpoint[ip, 0] = p_row + 0
            goodpoint[ip, 1] = p_col - 1
            ip = ip + 1
    # dir 6
    if BD_Out_Lakecat_Nriv_mask[p_row - 1, p_col - 1] == lid:

        tt = goodpoint[
            goodpoint[:, 0] == p_row - 1,
        ]
        if (
            len(
                tt[
                    tt[:, 1] == p_col - 1,
                ]
            )
            < 1
        ):
            ndir[p_row - 1, p_col - 1] = 2
            changed_ndir[p_row - 1, p_col - 1] = 2
            goodpoint[ip, 0] = p_row - 1
            goodpoint[ip, 1] = p_col - 1
            ip = ip + 1
    # dir 7
    if BD_Out_Lakecat_Nriv_mask[p_row - 1, p_col + 0] == lid:

        tt = goodpoint[
            goodpoint[:, 0] == p_row - 1,
        ]
        if (
            len(
                tt[
                    tt[:, 1] == p_col + 0,
                ]
            )
            < 1
        ):
            ndir[p_row - 1, p_col + 0] = 4
            changed_ndir[p_row - 1, p_col + 0] = 4
            goodpoint[ip, 0] = p_row - 1
            goodpoint[ip, 1] = p_col + 0
            ip = ip + 1
    # dir 8
    if BD_Out_Lakecat_Nriv_mask[p_row - 1, p_col + 1] == lid:

        tt = goodpoint[
            goodpoint[:, 0] == p_row - 1,
        ]
        if (
            len(
                tt[
                    tt[:, 1] == p_col + 1,
                ]
            )
            < 1
        ):
            ndir[p_row - 1, p_col + 1] = 8
            changed_ndir[p_row - 1, p_col + 1] = 8
            goodpoint[ip, 0] = p_row - 1
            goodpoint[ip, 1] = p_col + 1
            ip = ip + 1
    #    arcpy.AddMessage(goodpoint[goodpoint[:,0]>0,])
    return ndir, goodpoint, ip, changed_ndir


def is_point_in_2d_array(array, value_x, value_y):
    vx_is_in = np.isin(array[:, 0], value_x)
    vy_is_in = np.isin(array[:, 1], value_y)
    is_in = np.logical_and(vx_is_in, vy_is_in)
    # print("b")
    # print(array)
    # print(value_x,value_y)
    # print(vx_is_in)
    # print(vy_is_in)
    # print(is_in)
    # print(True in is_in)
    return (True in is_in, is_in)


def modify_flow_direction_of_nearby_boundary_grids(
    N_dir,
    p_row,
    p_col,
    lake_cat_grids,
    lid,
    goodpoint_loop,
    all_good_points,
    idx_good_loop,
    idx_good_all,
    problem_points,
    BD_Out_Lakecat_Nriv_mask,
    Changed_ndir,
):
    ### this function change flow direction of some grids around p_row, p_col, flow to p_row, p_col

    ndir = copy.copy(N_dir)
    changed_ndir = copy.copy(Changed_ndir)
    lake_cat_grids_new = copy.copy(lake_cat_grids)
    problem_points_new = copy.copy(problem_points)
    all_good_points_new = copy.copy(all_good_points)
    goodpoint_loop_new = copy.copy(goodpoint_loop)
    idx_good_loop_n = copy.copy(idx_good_loop)
    idx_good_all_n = copy.copy(idx_good_all)

    # dir 1
    #### it is a broudary grids and did not flow to the lake outlet
    if BD_Out_Lakecat_Nriv_mask[p_row + 0, p_col + 1] == lid:
        # check if it already modified
        is_alreay_in_good_points, temp = is_point_in_2d_array(
            all_good_points, p_row + 0, p_col + 1
        )
        #### the point not exist in good points which store all boundary points
        if not is_alreay_in_good_points:
            # print("a",is_alreay_in_good_points, not is_alreay_in_good_points)
            # print(all_good_points)
            # print(p_row + 0,p_col + 1)
            ### change the flow direction of new point to old points
            ndir[p_row + 0, p_col + 1] = 16
            changed_ndir[p_row + 0, p_col + 1] = 16
            goodpoint_loop_new[idx_good_loop_n, 0] = p_row + 0
            goodpoint_loop_new[idx_good_loop_n, 1] = p_col + 1
            idx_good_loop_n = idx_good_loop_n + 1
            all_good_points_new[idx_good_all_n, 0] = p_row + 0
            all_good_points_new[idx_good_all_n, 1] = p_col + 1
            idx_good_all_n = idx_good_all_n + 1
            # remote this point from good point
            is_alreay_in_pm_points, row_of_problempoints = is_point_in_2d_array(
                problem_points_new, p_row + 0, p_col + 1
            )
            problem_points_new = problem_points_new[
                np.logical_not(row_of_problempoints)
            ]
            # print("a0")
            # print(len(row_of_problempoints),len(problem_points_new),np.sum(row_of_problempoints))
            # print(is_alreay_in_good_points)
            # print("a0")
            # modify lake cat grids to 1
            lake_cat_grids_new[p_row + 0, p_col + 1] = 1

    # dir 2
    if BD_Out_Lakecat_Nriv_mask[p_row + 1, p_col + 1] == lid:

        is_alreay_in_good_points, temp = is_point_in_2d_array(
            all_good_points, p_row + 1, p_col + 1
        )
        #### the point not exist in good points which store all boundary points
        if not is_alreay_in_good_points:
            ndir[p_row + 1, p_col + 1] = 32
            changed_ndir[p_row + 1, p_col + 1] = 32
            goodpoint_loop_new[idx_good_loop_n, 0] = p_row + 1
            goodpoint_loop_new[idx_good_loop_n, 1] = p_col + 1
            idx_good_loop_n = idx_good_loop_n + 1
            all_good_points_new[idx_good_all_n, 0] = p_row + 1
            all_good_points_new[idx_good_all_n, 1] = p_col + 1
            idx_good_all_n = idx_good_all_n + 1
            # remote this point from good point
            is_alreay_in_pm_points, row_of_problempoints = is_point_in_2d_array(
                problem_points_new, p_row + 1, p_col + 1
            )
            problem_points_new = problem_points_new[
                np.logical_not(row_of_problempoints)
            ]
            # print("a1")
            # print(len(row_of_problempoints),len(problem_points_new),np.sum(row_of_problempoints))
            # print(is_alreay_in_good_points)
            # print("a1")

            # modify lake cat grids to 1
            lake_cat_grids_new[p_row + 1, p_col + 1] = 1

    # dir 3
    if BD_Out_Lakecat_Nriv_mask[p_row + 1, p_col + 0] == lid:

        is_alreay_in_good_points, temp = is_point_in_2d_array(
            all_good_points, p_row + 1, p_col + 0
        )
        #### the point not exist in good points which store all boundary points
        if not is_alreay_in_good_points:
            ndir[p_row + 1, p_col + 0] = 64
            changed_ndir[p_row + 1, p_col + 0] = 64
            goodpoint_loop_new[idx_good_loop_n, 0] = p_row + 1
            goodpoint_loop_new[idx_good_loop_n, 1] = p_col + 0
            idx_good_loop_n = idx_good_loop_n + 1
            all_good_points_new[idx_good_all_n, 0] = p_row + 1
            all_good_points_new[idx_good_all_n, 1] = p_col + 0
            idx_good_all_n = idx_good_all_n + 1
            # remote this point from good point
            is_alreay_in_pm_points, row_of_problempoints = is_point_in_2d_array(
                problem_points_new, p_row + 1, p_col + 0
            )
            problem_points_new = problem_points_new[
                np.logical_not(row_of_problempoints)
            ]
            # print("a2")
            # print(len(row_of_problempoints),len(problem_points_new),np.sum(row_of_problempoints))
            # print(is_alreay_in_good_points)
            # print("a2")

            # modify lake cat grids to 1
            lake_cat_grids_new[p_row + 1, p_col + 0] = 1

    # dir 4
    if BD_Out_Lakecat_Nriv_mask[p_row + 1, p_col - 1] == lid:
        is_alreay_in_good_points, temp = is_point_in_2d_array(
            all_good_points, p_row + 1, p_col - 1
        )
        #### the point not exist in good points which store all boundary points
        if not is_alreay_in_good_points:
            ndir[p_row + 1, p_col - 1] = 128
            changed_ndir[p_row + 1, p_col - 1] = 128
            goodpoint_loop_new[idx_good_loop_n, 0] = p_row + 1
            goodpoint_loop_new[idx_good_loop_n, 1] = p_col - 1
            idx_good_loop_n = idx_good_loop_n + 1
            all_good_points_new[idx_good_all_n, 0] = p_row + 1
            all_good_points_new[idx_good_all_n, 1] = p_col - 1
            idx_good_all_n = idx_good_all_n + 1
            # remote this point from good point
            is_alreay_in_pm_points, row_of_problempoints = is_point_in_2d_array(
                problem_points_new, p_row + 1, p_col - 1
            )
            problem_points_new = problem_points_new[
                np.logical_not(row_of_problempoints)
            ]
            # print("a3")
            # print(len(row_of_problempoints),len(problem_points_new),np.sum(row_of_problempoints))
            # print(is_alreay_in_good_points)
            # print("a3")

            # modify lake cat grids to 1
            lake_cat_grids_new[p_row + 1, p_col - 1] = 1

    # dir 5
    if BD_Out_Lakecat_Nriv_mask[p_row + 0, p_col - 1] == lid:

        is_alreay_in_good_points, temp = is_point_in_2d_array(
            all_good_points, p_row + 0, p_col - 1
        )
        #### the point not exist in good points which store all boundary points
        if not is_alreay_in_good_points:
            ndir[p_row + 0, p_col - 1] = 1
            changed_ndir[p_row + 0, p_col - 1] = 1
            goodpoint_loop_new[idx_good_loop_n, 0] = p_row + 0
            goodpoint_loop_new[idx_good_loop_n, 1] = p_col - 1
            idx_good_loop_n = idx_good_loop_n + 1
            all_good_points_new[idx_good_all_n, 0] = p_row + 0
            all_good_points_new[idx_good_all_n, 1] = p_col - 1
            idx_good_all_n = idx_good_all_n + 1
            # remote this point from good point
            is_alreay_in_pm_points, row_of_problempoints = is_point_in_2d_array(
                problem_points_new, p_row + 0, p_col - 1
            )
            problem_points_new = problem_points_new[
                np.logical_not(row_of_problempoints)
            ]
            # print("a4")
            # print(len(row_of_problempoints),len(problem_points_new),np.sum(row_of_problempoints))
            # print(is_alreay_in_good_points)
            # print("a4")

            # modify lake cat grids to 1
            lake_cat_grids_new[p_row + 0, p_col - 1] = 1

    # dir 6
    if BD_Out_Lakecat_Nriv_mask[p_row - 1, p_col - 1] == lid:

        is_alreay_in_good_points, temp = is_point_in_2d_array(
            all_good_points, p_row - 1, p_col - 1
        )
        #### the point not exist in good points which store all boundary points
        if not is_alreay_in_good_points:
            ndir[p_row - 1, p_col - 1] = 2
            changed_ndir[p_row - 1, p_col - 1] = 2
            goodpoint_loop_new[idx_good_loop_n, 0] = p_row - 1
            goodpoint_loop_new[idx_good_loop_n, 1] = p_col - 1
            idx_good_loop_n = idx_good_loop_n + 1
            all_good_points_new[idx_good_all_n, 0] = p_row - 1
            all_good_points_new[idx_good_all_n, 1] = p_col - 1
            idx_good_all_n = idx_good_all_n + 1
            # remote this point from good point
            is_alreay_in_pm_points, row_of_problempoints = is_point_in_2d_array(
                problem_points_new, p_row - 1, p_col - 1
            )
            problem_points_new = problem_points_new[
                np.logical_not(row_of_problempoints)
            ]
            # print("a5")
            # print(len(row_of_problempoints),len(problem_points_new),np.sum(row_of_problempoints))
            # print(is_alreay_in_good_points)
            # print("a5")

            # modify lake cat grids to 1
            lake_cat_grids_new[p_row - 1, p_col - 1] = 1

    # dir 7
    if BD_Out_Lakecat_Nriv_mask[p_row - 1, p_col + 0] == lid:

        is_alreay_in_good_points, temp = is_point_in_2d_array(
            all_good_points, p_row - 1, p_col + 0
        )
        #### the point not exist in good points which store all boundary points
        if not is_alreay_in_good_points:
            ndir[p_row - 1, p_col + 0] = 4
            changed_ndir[p_row - 1, p_col + 0] = 4
            goodpoint_loop_new[idx_good_loop_n, 0] = p_row - 1
            goodpoint_loop_new[idx_good_loop_n, 1] = p_col + 0
            idx_good_loop_n = idx_good_loop_n + 1
            all_good_points_new[idx_good_all_n, 0] = p_row - 1
            all_good_points_new[idx_good_all_n, 1] = p_col + 0
            idx_good_all_n = idx_good_all_n + 1
            # remote this point from good point
            is_alreay_in_pm_points, row_of_problempoints = is_point_in_2d_array(
                problem_points_new, p_row - 1, p_col + 0
            )
            problem_points_new = problem_points_new[
                np.logical_not(row_of_problempoints)
            ]
            # print("a6")
            # print(len(row_of_problempoints),len(problem_points_new),np.sum(row_of_problempoints))
            # print(is_alreay_in_good_points)
            # print("a6")

            # modify lake cat grids to 1
            lake_cat_grids_new[p_row - 1, p_col + 0] = 1

    # dir 8
    if BD_Out_Lakecat_Nriv_mask[p_row - 1, p_col + 1] == lid:

        is_alreay_in_good_points, temp = is_point_in_2d_array(
            all_good_points, p_row - 1, p_col + 1
        )
        #### the point not exist in good points which store all boundary points
        if not is_alreay_in_good_points:
            ndir[p_row - 1, p_col + 1] = 8
            changed_ndir[p_row - 1, p_col + 1] = 8
            goodpoint_loop_new[idx_good_loop_n, 0] = p_row - 1
            goodpoint_loop_new[idx_good_loop_n, 1] = p_col + 1
            idx_good_loop_n = idx_good_loop_n + 1
            all_good_points_new[idx_good_all_n, 0] = p_row - 1
            all_good_points_new[idx_good_all_n, 1] = p_col + 1
            idx_good_all_n = idx_good_all_n + 1
            # remote this point from good point
            is_alreay_in_pm_points, row_of_problempoints = is_point_in_2d_array(
                problem_points_new, p_row - 1, p_col + 1
            )
            problem_points_new = problem_points_new[
                np.logical_not(row_of_problempoints)
            ]
            # print("a7")
            # print(len(row_of_problempoints),len(problem_points_new),np.sum(row_of_problempoints))
            # print(is_alreay_in_good_points)
            # print("a7")

            # modify lake cat grids to 1
            lake_cat_grids_new[p_row - 1, p_col + 1] = 1

    #    arcpy.AddMessage(goodpoint[goodpoint[:,0]>0,])
    return (
        ndir,
        changed_ndir,
        goodpoint_loop_new,
        idx_good_loop_n,
        all_good_points_new,
        idx_good_all_n,
        problem_points_new,
        lake_cat_grids_new,
    )



def watershed_delineation_without_lake(
    acc_thresold,
    mode,
    input_geo_names,
    max_memroy=1024 * 4,
    grassdb="#",
    grass_location="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
    fdr_path="#",
    fac_path = '#',
    subreg_fdr_path="#",
    subreg_acc_path="#",
    subreg_str_r_path="#",
    subreg_str_v_path="#",
    fdr_arcgis="fdr_arcgis",
    fdr_grass="fdr_grass",
    str_r="str_r",
    str_v="str_v",
    acc="acc",
    cat_no_lake="cat_no_lake",
):
    """Generate a subbasin delineation without considering lake

    Function that used to Generate a subbasin delineation and river
    network using user provied flow accumulation thresthold
    without considering lake.

    Parameters
    ----------
    accthresold       : float
        It is the flow accumulation thresthold, used to determine
        subbsains and river network. Increasing of accthresold will
        increase the size of generated subbasins, reduce the number
        subbasins and reduce the number of generated stream segments
    mode              : string (required)
        It is a string indicate which dataset will be used to delineate
        watershed.
        'usingdem'             : dem is used for delineation
        'usingfdr'             : flow direction data is used for delineation
        'usingsubreg'          : predefined subregion inputs is used for
    input_geo_names    : dict
        it is a dictionary that list the required input file names,should at
        least indicate the name of following items:

        mask                   : raster
            it is a mask raster stored in grass database, which indicate
            the PSE. The grass database is located at
            os.path.join(grassdb, grass_location)
        dem                   : raster
            it is a dem raster stored in grass database, which is
            has the same extent with MASK. The grass database is located at
            os.path.join(grassdb, grass_location)
    max_memroy        : integer
        It is the maximum memeory that allow to be used.
    grassdb           : path (required)
        It is a path to project grass database folder
    grass_location    : string (required)
        It is a string of grass location name
    qgis_prefix_path  : string (required)
        It is a string of qgis prefix path
    gis_platform      : string
        It is a string indicate with gis platform is used:
        'qgis'                : the basinmaker is running within QGIS
        'arcgis'              : the basinmaker is running within ArcGIS
    fdr_path          : string
        It is a string indicate path of flow direction dataset
    subreg_fdr_path   : string
        It is a string indicate path of subregion flow direction dataset
    subreg_acc_path   : string
        It is a string indicate path of subregion flow accumlation dataset
    subreg_str_r_path : string
        It is a string indicate path of subregion stream raster flow
        direction dataset
    subreg_str_v_path : string
        It is a string indicate path of subregion stream vector datasets

    Notes
    -------
    Outputs are following files

    fdr_grass              : raster
        it is a raster represent flow direction dataset, which is
        using 1 - 8 to represent different directions
    fdr_arcgis             : raster
        it is a raster represent flow direction dataset, which is
        using 1,2,4,...64,128 to represent different directions
    str_v                  : vector
        it is a river network in vector format
    str_r                  : raster
        it is a river network in raster format
    cat_no_lake            : raster
         it is the raster represent the delineated subbasins without
         considering lakes
    acc                    : raster
         it is the raster represent the flow accumulation

    Returns:
    -------
       None

    Examples
    -------
    """

    if gis_platform == "qgis":
        assert (
            grassdb != "#"
        ), "grass database folder is needed, when gis_platform = qgis"
        assert (
            grass_location != "#"
        ), "grass location name is needed, when gis_platform = qgis"
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.delineationnolake.watusingdemqgis import (
            delineate_watershed_no_lake_using_dem,
        )
        from basinmaker.delineationnolake.watusingfdrqgis import (
            delineate_watershed_no_lake_using_fdr,
        )
        from basinmaker.delineationnolake.watusingfacqgis import (
            delineate_watershed_no_lake_using_fac,
        )
        from basinmaker.delineationnolake.watusingsubregionddata import (
            delineate_watershed_no_lake_using_subregion_data,
        )
    elif gis_platform == "arcgis":
        from basinmaker.delineationnolake.watusingdemarcgis import (
            delineate_watershed_no_lake_using_dem,
        )
        from basinmaker.delineationnolake.watusingfdrarcgis import (
            delineate_watershed_no_lake_using_fdr,
        )
    else:
        print('gis_platform    ',gis_platform,'   do not support')    
        
    if mode == "usingdem":
        assert (
            acc_thresold > 0
        ), "the flow accumulation thresthold is needed to delineate watershed from dem"

    if mode == "usingfdr":
        assert (
            fdr_path != "#"
        ), "The path of the provided flow direction data is needed to delineate watershed from flow direction"

    if mode == "usingfac":
        assert (
            fac_path != "#"
        ), "The path of the provided flow direction data is needed to delineate watershed from flow direction"
        
    if mode == "usingsubreg":
        assert (
            subreg_acc_path != "#"
        ), "subregion flow accumulation is neeeded to delineate watershed for current subregion "

        assert (
            subreg_fdr_path != "#"
        ), "subregion flow direction is neeeded to delineate watershed for current subregion "

        assert (
            subreg_str_r_path != "#"
        ), "subregion stream raster is neeeded to delineate watershed for current subregion "

        assert (
            subreg_str_v_path != "#"
        ), "subregion stream vector is neeeded to delineate watershed for current subregion "

    if mode == "usingdem":
        delineate_watershed_no_lake_using_dem(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            input_geo_names=input_geo_names,
            acc_thresold=acc_thresold,
            fdr_arcgis=fdr_arcgis,
            fdr_grass=fdr_grass,
            str_r=str_r,
            str_v=str_v,
            acc=acc,
            cat_no_lake=cat_no_lake,
            max_memroy=max_memroy,
        )

    if mode == "usingfdr":
        delineate_watershed_no_lake_using_fdr(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            input_geo_names=input_geo_names,
            fdr_path=fdr_path,
            acc_thresold=acc_thresold,
            fdr_arcgis=fdr_arcgis,
            fdr_grass=fdr_grass,
            str_r=str_r,
            str_v=str_v,
            acc=acc,
            cat_no_lake=cat_no_lake,
            max_memroy=max_memroy,
        )
    if mode == "usingfac":
        delineate_watershed_no_lake_using_fac(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            input_geo_names=input_geo_names,
            fac_path=fac_path,
            acc_thresold=acc_thresold,
            fdr_arcgis=fdr_arcgis,
            fdr_grass=fdr_grass,            
            str_r=str_r,
            str_v=str_v,
            acc=acc,
            cat_no_lake=cat_no_lake,
            max_memroy=max_memroy,
        )        
    if mode == "usingsubreg":
        delineate_watershed_no_lake_using_subregion_data(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            input_geo_names=input_geo_names,
            subreg_fdr_path=subreg_fdr_path,
            subreg_acc_path=subreg_acc_path,
            subreg_str_v_path=subreg_str_v_path,
            subreg_str_r_path=subreg_str_r_path,
            fdr_arcgis=fdr_arcgis,
            fdr_grass=fdr_grass,
            str_r=str_r,
            str_v=str_v,
            acc=acc,
            cat_no_lake=cat_no_lake,
            max_memroy=max_memroy,
        )



import os

from simpledbf import Dbf5

# define internal file names
Internal_Constant_Names = {
    "cat_add_lake_old_fdr": "cat_add_lake_old_fdr",
    "cat_add_lake": "cat_add_lake",
    "pourpoints_add_obs": "pourpoints_add_obs",
    "all_lakes": "all_lakes",
    "lake_boundary": "lake_boundary",
    "connect_lake": "connect_lake",
    "nonconnect_lake": "nonconnect_lake",
    "str_connected_lake": "str_connected_lake",
    "str_sl_connected_lake": "str_sl_connected_lake",
    "nfdr_arcgis": "nfdr_arcgis",
    "nfdr_grass": "nfdr_grass",
    "obs": "obs",
    "pourpoints_with_lakes": "pourpoints_with_lakes",
    "lake_inflow_pourpoints": "lake_inflow_pourpoints",
    "lake_outflow_pourpoints": "lake_outflow_pourpoints",
    "catchment_pourpoints_outside_lake": "catchment_pourpoints_outside_lake",
    "cat_use_default_acc": "cat_use_default_acc",
    "cat_ply_info": "cat_ply_info",
    "cat_riv_info": "cat_riv_info",
    "outlet_pt_info": "outlet_pt_info",
    "fdr_grass": "fdr_grass",
    "fdr_arcgis": "fdr_arcgis",
    "str_r": "str_r",
    "str_v": "str_v",
    "acc": "acc",
    "cat_no_lake": "cat_no_lake",
    "sl_connected_lake": "sl_connected_lake",
    "sl_nonconnect_lake": "sl_nonconnect_lake",
    "selected_lakes": "selected_lakes",
    "catchment_without_merging_lakes": "catchment_without_merging_lakes",
    "river_without_merging_lakes": "river_without_merging_lakes",
    "snapped_obs_points": "snapped_obs_points",
    "cat_use_default_acc": "cat_use_default_acc",
    "snapped_obs_points": "snapped_obs_points",
}


####
COLUMN_NAMES_CONSTANT = [
    "SubId",
    "DowSubId",
    "RivSlope",
    "RivLength",
    "BasSlope",
    "BasAspect",
    "BasArea",
    "BkfWidth",
    "BkfDepth",
    "Lake_Cat",
    "HyLakeId",
    "LakeVol",
    "LakeDepth",
    "LakeArea",
    "Laketype",
    "Has_POI",
    "MeanElev",
    "FloodP_n",
    "Q_Mean",
    "Ch_n",
    "DrainArea",
    "Strahler",
    "Seg_ID",
    "Seg_order",
    "Max_DEM",
    "Min_DEM",
    "DA_Obs",
    "DA_error",
    "Obs_NM",
    "SRC_obs",
    "centroid_x",
    "centroid_y",
    "DA_Chn_L",
    "DA_Slope",
    "DA_Chn_Slp",
    "outletLat",
    "outletLng",
]

COLUMN_NAMES_CONSTANT_CLEAN = [
    "SubId",
    "DowSubId",
    "RivSlope",
    "RivLength",
    "BasSlope",
    "BasAspect",
    "BasArea",
    "BkfWidth",
    "BkfDepth",
    "Lake_Cat",
    "HyLakeId",
    "LakeVol",
    "LakeDepth",
    "LakeArea",
    "Laketype",
    "Has_POI",
    "MeanElev",
    "FloodP_n",
    "Q_Mean",
    "Ch_n",
    "DrainArea",
    "Strahler",
    "Seg_ID",
    "Seg_order",
    "Max_DEM",
    "Min_DEM",
    "DA_Obs",
    "DA_error",
    "Obs_NM",
    "SRC_obs",
    "centroid_x",
    "centroid_y",
    "DA_Chn_L",
    "DA_Slope",
    "DA_Chn_Slp",
    "outletLat",
    "outletLng",
    "Has_Gauge"
]


COLUMN_NAMES_With_NULL_Values = [
    "Lake_Cat",
    "HyLakeId",
    "LakeVol",
    "LakeDepth",
    "LakeArea",
    "Laketype",
    "DA_Obs",
    "DA_error",
    "Obs_NM",
    "SRC_obs",
]

COLUMN_NAMES_CONSTANT_HRU = [
    "SubId",
    "DowSubId",
    "RivSlope",
    "RivLength",
    "BasSlope",
    "BasAspect",
    "BasArea",
    "BkfWidth",
    "BkfDepth",
    "Lake_Cat",
    "HyLakeId",
    "LakeVol",
    "LakeDepth",
    "LakeArea",
    "Laketype",
    "Has_POI",
    "MeanElev",
    "FloodP_n",
    "Q_Mean",
    "Ch_n",
    "DrainArea",
    "Strahler",
    "Seg_ID",
    "Seg_order",
    "Max_DEM",
    "Min_DEM",
    "DA_Obs",
    "DA_error",
    "Obs_NM",
    "SRC_obs",
    "centroid_x",
    "centroid_y",
    "HRU_IsLake",
    "Landuse_ID",
    "Soil_ID",
    "Veg_ID",
    "O_ID_1",
    "O_ID_2",
    "HRU_Area",
    "HRU_ID",
    "LAND_USE_C",
    "VEG_C",
    "SOIL_PROF",
    "HRU_CenX",
    "HRU_CenY",
    "HRU_S_mean",
    "HRU_A_mean",
    "HRU_E_mean",
    "SHAPE",
    "DA_Chn_L",
    "DA_Slope",
    "DA_Chn_Slp",
    "Has_Gauge" ,   
]


COLUMN_TYPES_CONSTANT = [
    "Integer",
    "Integer",
    "Real",
    "Real",
    "Real",
    "Real",
    "Real",
    "Real",
    "Real",
    "Integer",
    "Integer",
    "Real",
    "Real",
    "Real",
    "Integer",
    "Integer",
    "Real",
    "Real",
    "Real",
    "Real",
    "Real",
    "Integer",
    "Integer",
    "Integer",
    "Real",
    "Real",
    "Real",
    "Real",
    "Character",
    "Character",
    "Real",
    "Real",
    "Real",
    "Real",
    "Real",
    "Real",
    "Real",
]

DEFALUT_FLOOD_N = 0.09
min_manning_n = 0.025
max_manning_n = 0.15
min_riv_slope = 0.000001
max_riv_slope = 1
min_bkf_width = 0.1
min_bkf_depth = 0.1
min_riv_lenth = 90
    
def Dbf_To_Dataframe(file_path):
    """Transfer an input dbf file to dataframe

    Parameters
    ----------
    file_path   : string
    Full path to a shapefile

    Returns:
    -------
    dataframe   : datafame
    a pandas dataframe of attribute table of input shapefile
    """
    tempinfo = Dbf5(file_path[:-3] + "dbf")
    dataframe = tempinfo.to_dataframe().copy()
    return dataframe


def WriteStringToFile(Out_String, File_Path, WriteMethod):
    """Write String to a file

    Function that used to write Out_String to a file located at the File_Path.

    Parameters
    ----------
    Out_String            : string
        The string that will be writed to the file located at File_Path
    File_Path             : string
        Path and filename of file that will be modified or created
    WriteMethod           : {'a','w'}
        If WriteMethod = "w", a new file will be created at the File_Path
        If WriteMethod = "a", the Out_String will be added to exist file

    Notes
    ------
        The file located at the File_Path will be modified or created

    Returns
    -------
        None

    Examples
    --------
    >>> from WriteRavenInputs import WriteStringToFile
    >>> Out_String = 'sometest at line 1\n some test at line 2\n some test at line 3\n'
    >>> File_Path  = 'C:/Path_to_the_Flie_with_file_name'
    >>> WriteStringToFile(Out_String = Out_String,File_Path = File_Path,WriteMethod = 'w')

    """

    if os.path.exists(
        File_Path
    ):  ### if file exist, we can either modify or overwrite it
        with open(File_Path, WriteMethod) as f:
            f.write(Out_String)
    else:  ## create a new file anyway, since file did not exist
        with open(File_Path, "w") as f:
            f.write(Out_String)


def write_grass_and_arcgis_fdr_rules(grassdb):

    Strlist = [
        "1 = 128",
        "2 = 64",
        "3 = 32",
        "4 = 16",
        "5 = 8",
        "6 = 4",
        "7 = 2",
        "8 = 1",
        "* = NULL",
    ]
    Str = "\n".join(Strlist)
    WriteStringToFile(Str, os.path.join(grassdb, "Grass2ArcgisDIR.txt"), "w")

    Strlist = [
        "1 = 8",
        "2 = 7",
        "4 = 6",
        "8 = 5",
        "16 = 4",
        "32 = 3",
        "64 = 2",
        "128 = 1",
        "* = NULL",
    ]
    Str = "\n".join(Strlist)
    WriteStringToFile(Str, os.path.join(grassdb, "Arcgis2GrassDIR.txt"), "w")

    return


def write_grass_reclass_rule_from_table(table, output_path):
    Strlist = []
    for i in range(0, len(table)):
        Strlist.append(str(table[i, 0]) + " = " + str(table[i, 1]))
    Strlist.append("* = NULL")
    Str = "\n".join(Strlist)
    WriteStringToFile(Str, output_path, "w")



import os
import geopandas
from basinmaker.func.pdtable import *
import numpy as np
from basinmaker.postprocessing.postprocessingfunctions import (
    combine_catchments_covered_by_the_same_lake_method,
)

def inverse_topology_subids(datafame,end_subid,start_subid):
    datafame['ndownsubid'] =datafame['DowSubId']
    csubid = end_subid
    up_subid = -1

    false_drainage_csubid = datafame.loc[datafame['DowSubId'] == csubid,'SubId'].values
#    print(len(false_drainage_csubid))
    if len(false_drainage_csubid) > 2:
        cmask = datafame['SubId'] == csubid
    else:
        false_drainage_csubid = false_drainage_csubid[false_drainage_csubid != up_subid]
        mask_subid = np.append(false_drainage_csubid,[csubid])
        cmask = datafame['SubId'].isin(mask_subid)

#    print(csubid,datafame.loc[cmask,'DowSubId'].values[0])

    while csubid != -1:
        datafame.loc[cmask,'ndownsubid'] = up_subid
        up_subid = csubid
        csubid = datafame.loc[datafame['SubId'] == csubid,'DowSubId'].values[0]

        false_drainage_csubid = datafame.loc[datafame['DowSubId'] == csubid,'SubId'].values
        if len(false_drainage_csubid) > 2:
            cmask = datafame['SubId'] == csubid
        else:
            false_drainage_csubid = false_drainage_csubid[false_drainage_csubid != up_subid]
            mask_subid = np.append(false_drainage_csubid,[csubid])
            cmask = datafame['SubId'].isin(mask_subid)

#        print(csubid,up_subid,datafame.loc[cmask,'SubId'].values)

    datafame['DowSubId'] = datafame['ndownsubid']
    datafame = datafame.drop(columns=['ndownsubid'])

    datafame = streamorderanddrainagearea(datafame)

    datafame = update_non_connected_catchment_info(datafame)
    datafame.loc[datafame['RivLength'] == -1.2345,'RivSlope'] = -1.2345
    datafame.loc[datafame['RivLength'] == -1.2345,'FloodP_n'] = -1.2345
    datafame.loc[datafame['RivLength'] == -1.2345,'Max_DEM'] = -1.2345
    datafame.loc[datafame['RivLength'] == -1.2345,'Min_DEM'] = -1.2345
    datafame.loc[datafame['RivLength'] == -1.2345,'Ch_n'] = -1.2345


    return datafame


def inverse_topology(path_to_product_folder,path_to_output_folder,version_number,end_subid,start_subid):

    if not os.path.exists(path_to_output_folder):
        os.mkdir(path_to_output_folder)

    path_subbasin = os.path.join(path_to_product_folder,'catchment_without_merging_lakes_'+version_number+'.shp')
    path_river = os.path.join(path_to_product_folder,'river_without_merging_lakes_'+version_number+'.shp')

    subbasin = geopandas.read_file(path_subbasin)

    river = geopandas.read_file(path_river)
    river = river[['SubId','geometry']]
    subbasin = inverse_topology_subids(subbasin,end_subid,start_subid)
    river = pd.merge(river, subbasin.drop(columns=['geometry']), on = 'SubId', how='inner')

    subbasin.to_file(os.path.join(path_to_output_folder,'catchment_without_merging_lakes_'+version_number+'.shp'))
    river.to_file(os.path.join(path_to_output_folder,'river_without_merging_lakes_'+version_number+'.shp'))

    combine_catchments_covered_by_the_same_lake_method(
        Routing_Product_Folder = path_to_output_folder,
        qgis_prefix_path='#',
        gis_platform='purepy',
    )



import geopandas
import numpy as np
import os
import pandas as pd
from json import load, JSONEncoder
import json
import requests
import sys
#import shapely.wkt
#import pygeos as pg
from osgeo import gdal, ogr
from rasterstats import zonal_stats

def zonal_stats_pd(shp_gpd,raster,stats,key):

    result = zonal_stats(shp_gpd, raster, stats = stats,geojson_out=True,all_touched=True)

    reault_list_dic = list(map(lambda x : {key:x['properties'][key],'mean':x['properties']['mean']}, result))

    reault_pd = pd.DataFrame(reault_list_dic)

    return reault_pd



def ZonalStats(shp_gpd, raster, stats,key,col):
    # shape - shapefile path
    # raster - raster path
    # stats - stats as list, f.e. 'min mean max' ; 'min'
    # the result is final_gdf as GeoDataFrame

    nodata_pd = shp_gpd.copy(deep=True)
    i = 0
    while len(nodata_pd) >0 and i <=5:
        temp_pd = zonal_stats_pd(nodata_pd,raster,stats,key)
        bad_hru_values = temp_pd[~(temp_pd['mean'] > 0)]['HRU_ID_New']
        nodata_pd = nodata_pd[nodata_pd['HRU_ID_New'].isin(bad_hru_values)].copy(deep=True)
        if i == 0:
            reault_pd = temp_pd[temp_pd['mean'] > 0]
            i = i + 1
        else:
            i = i + 1
            good_pd = temp_pd[temp_pd['mean'] > 0]
            reault_pd = reault_pd.append(good_pd)

    if len(nodata_pd) > 0:
        nodata_pd['mean'] = nodata_pd[col]
        reault_pd = reault_pd.append(nodata_pd)
#    print(len(shp_gpd),len(nodata_pd),len(reault_pd))
    return reault_pd


def Reproj_Clip_Dissolve_Simplify_Polygon_purepy(
    layer_path, Class_Col, tempfolder,mask_layer,Class_NM_Col = '#',info_table = '#'
):
    """Preprocess user provided polygons

    Function that will reproject clip input polygon with subbasin polygon
    and will dissolve the input polygon based on their ID, such as landuse id
    or soil id.

    Parameters
    ----------
    processing                        : qgis object
    context                           : qgis object
    layer_path                        : string
        The path to a specific polygon, for example path to landuse layer
    Project_crs                       : string
        the EPSG code of a projected coodinate system that will be used to
        calcuate HRU area and slope.
    trg_crs                           : string
        the EPSG code of a  coodinate system that will be used to
        calcuate reproject input polygon
    Class_Col                         : string
        the column name in the input polygon (layer_path) that contains
        their ID, for example land use ID or soil ID.
    Layer_clip                        : qgis object
        A shpfile with extent of the watershed, will be used to clip input
        input polygon
    Notes
    -------
        # TODO: May be add some function to simplify the input polygons
                for example, remove the landuse type with small areas
                or merge small landuse polygon into the surrounding polygon

    Returns:
    -------
        layer_dis                  : qgis object
            it is a polygon after preprocess
    """

    new_crs = mask_layer.crs
    data = geopandas.read_file(layer_path)

    projected = data.to_crs(new_crs)
    projected = projected.explode(ignore_index=True)
    clipped = projected.clip(mask_layer)
#    dissolved = clipped.dissolve(by=[Class_Col], aggfunc='first',as_index=False)
#    print(Class_Col,new_crs,clipped.crs)
    cleaned = clean_geometry_purepy(clipped,1)
    if Class_NM_Col != '#':
        if Class_Col not in clipped.columns:
            print(Class_Col," not in the attribute table of provided shapefile")
            sys.exit()
        if Class_Col not in info_table.columns:
            print(Class_Col," not in the attribute table of provided info table")
            sys.exit()
        if Class_NM_Col not in info_table.columns:
            print(Class_NM_Col," not in the attribute table of provided info table")
            sys.exit()

        info_table_copy = info_table.copy(deep=True)
        info_table_copy = info_table_copy.drop_duplicates(subset=[Class_NM_Col], keep='last',ignore_index=True)
        info_table_copy['New_ID'] = info_table_copy[Class_Col]
        info_table_copy = info_table_copy[[Class_NM_Col,'New_ID']]
        info_table_copy2 = pd.merge(info_table, info_table_copy, how='inner', on = Class_NM_Col).copy(deep=True)

        cleaned = pd.merge(cleaned, info_table_copy2, how='inner', on = Class_Col)
        cleaned[Class_Col] = cleaned['New_ID']
        #update
    #    clipped Class_Col based on the Class_NM_Col
    #    arcpy.RepairGeometry_management(os.path.join(tempfolder,Class_Col+"_dislve.shp"))

    #    arcpy.AddSpatialIndex_management(os.path.join(tempfolder,Class_Col+"_dislve.shp"))
    return cleaned


def save_modified_attributes_to_outputs(mapoldnew_info,tempfolder,OutputFolder,cat_name,riv_name,Path_final_riv,dis_col_name='SubId'):


    NEED_TO_REMOVE_IDS = ["SubId_1", "Id","nsubid2", "nsubid","ndownsubid","Old_SubId","Old_DowSub","Join_Count","TARGET_FID","Id","SubID_Oldr","HRU_ID_N_1","HRU_ID_N_2","facters","Old_DowSubId","SubIdt2"]

    #obtain readme file
    if "DA_Chn_L" in  mapoldnew_info.columns:
        url = 'https://github.com/dustming/RoutingTool/wiki/Files/README_OIH.pdf'
    else:
        url = 'https://github.com/dustming/RoutingTool/wiki/Files/README_NA.pdf'

    response = requests.get(url)
    with open(os.path.join(OutputFolder,"README.pdf"), 'wb') as f:
        f.write(response.content)

    if riv_name != '#':

        if Path_final_riv != '#':
            riv_pd = geopandas.read_file(Path_final_riv)
            riv_pd['Old_SubId'] = riv_pd['SubId']

        cat_pd = mapoldnew_info.drop(columns = 'geometry').copy(deep=True)
        # remove all columns
        if Path_final_riv != '#':
            riv_pd = riv_pd[['geometry','Old_SubId']]
            riv_pd = pd.merge(riv_pd, cat_pd, on='Old_SubId', how='left')
            riv_pd = riv_pd.dissolve(by=dis_col_name, aggfunc='first',as_index=False)


        mapoldnew_info = mapoldnew_info.dissolve(by=dis_col_name, aggfunc='first',as_index=False)
        mapoldnew_info = add_centroid_in_wgs84(mapoldnew_info,"centroid_x","centroid_y")

        cat_c_x_y = mapoldnew_info[["centroid_y","centroid_x"]].copy(deep=True)
        if Path_final_riv != '#':
            riv_pd = riv_pd.drop(columns = ["centroid_y","centroid_x"])
            riv_pd = riv_pd.join(cat_c_x_y)

        riv_pd_nncls_routing_info = mapoldnew_info[mapoldnew_info['Lake_Cat'] != 2][['SubId','DowSubId']].copy(deep=True)
        remove_channel = []
        for subid in riv_pd_nncls_routing_info['SubId'].values:
            if subid not in riv_pd_nncls_routing_info['DowSubId'].values:
                remove_channel.append(subid)
        if Path_final_riv != '#':
            riv_pd = riv_pd[~riv_pd.SubId.isin(remove_channel)]
            cat_colnms = riv_pd.columns
            drop_cat_colnms = cat_colnms[cat_colnms.isin(NEED_TO_REMOVE_IDS)]
            riv_pd = riv_pd.drop(columns=drop_cat_colnms)
            if len(riv_pd) > 0:
                riv_pd.to_file(os.path.join(OutputFolder,riv_name))

        mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'RivSlope'] = -1.2345
        mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'RivLength'] = -1.2345
        mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'FloodP_n'] = -1.2345
        mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'Ch_n'] = -1.2345
        mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'Max_DEM'] = -1.2345
        mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'Min_DEM'] = -1.2345

        cat_colnms = mapoldnew_info.columns
        drop_cat_colnms = cat_colnms[cat_colnms.isin(NEED_TO_REMOVE_IDS)]
        mapoldnew_info = mapoldnew_info.drop(columns=drop_cat_colnms)
        mapoldnew_info.to_file(os.path.join(OutputFolder,cat_name))

        create_geo_jason_file(os.path.join(OutputFolder,cat_name))

    else:

        mapoldnew_info = mapoldnew_info.dissolve(by=dis_col_name, aggfunc='first',as_index=False)

        if "centroid_y" in mapoldnew_info.columns:

            mapoldnew_info = add_centroid_in_wgs84(mapoldnew_info,"centroid_x","centroid_y")
            mapoldnew_info["SubId"] = mapoldnew_info.index
            riv_pd_nncls_routing_info = mapoldnew_info[mapoldnew_info['Lake_Cat'] != 2][['SubId','DowSubId']].copy(deep=True)
            remove_channel = []
            for subid in riv_pd_nncls_routing_info['SubId'].values:
                if subid not in riv_pd_nncls_routing_info['DowSubId'].values:
                    remove_channel.append(subid)

            mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'RivSlope'] = -1.2345
            mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'RivLength'] = -1.2345
            mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'FloodP_n'] = -1.2345
            mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'Ch_n'] = -1.2345
            mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'Max_DEM'] = -1.2345
            mapoldnew_info.loc[mapoldnew_info.SubId.isin(remove_channel),'Min_DEM'] = -1.2345


        cat_colnms = mapoldnew_info.columns
        drop_cat_colnms = cat_colnms[cat_colnms.isin(["SHAPE","SubId_1", "Id","nsubid2", "nsubid","ndownsubid","Old_DowSub","Join_Count","TARGET_FID","Id","SubID_Oldr","HRU_ID_N_1","HRU_ID_N_2","facters","Old_DowSubId"])]
        mapoldnew_info = mapoldnew_info.drop(columns=drop_cat_colnms)
        mapoldnew_info.to_file(os.path.join(OutputFolder,cat_name))
        return mapoldnew_info


def Remove_Unselected_Lake_Attribute_In_Finalcatinfo_purepy(finalcat_ply, Conn_Lake_Ids):
    """Functions will set lake id not in Conn_Lake_Ids to -1.2345 in attribute
        table of Path_Finalcatinfo
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    mask1 = np.logical_not(finalcat_ply['HyLakeId'].isin(Conn_Lake_Ids))
    mask2 = finalcat_ply['Lake_Cat'] != 2
    mask = np.logical_and(mask1,mask2)

    finalcat_ply.loc[mask,'HyLakeId'] = 0
    finalcat_ply.loc[mask,'LakeVol'] = 0
    finalcat_ply.loc[mask,'LakeArea'] = 0
    finalcat_ply.loc[mask,'LakeDepth'] = 0
    finalcat_ply.loc[mask,'Laketype'] =0
    finalcat_ply.loc[mask,'Lake_Cat'] = 0

    return finalcat_ply

def clean_attribute_name_purepy(table,names):
    remove_column_names = table.columns[np.logical_not(np.isin(table.columns,names))]
    table = table.drop(columns=remove_column_names)
    return table

def clean_geometry_purepy(data,set_precision = -1):

#    data["geometry"] = data["geometry"].apply(lambda x: shapely.wkt.loads(shapely.wkt.dumps(x, rounding_precision=4)))
    if set_precision > 0:
        data['geometry'] = data['geometry'].buffer(0.00000000001)

#        data.geometry = pg.set_precision(data.geometry.values.data, 1e-6)
#        data["geometry"] = data["geometry"].apply(lambda x: shapely.wkt.loads(shapely.wkt.dumps(x, rounding_precision=4)))
#        data['geometry'] = data['geometry'].buffer(0)
#    print("aaaa")
    narow = ~data['geometry'].isna()
#    print("a1",len(data.loc[narow]))
    emrow = ~data.is_empty
#    print("a2",len(data.loc[emrow]))
    arearow = data.area > 0
#    print("a3",len(data.loc[arearow]))
#    arevalid = data.is_valid
    row1 = np.logical_and(narow,emrow)
    rowselect = np.logical_and(arearow,row1)
#    print("a",len(data))
#    rowselect = np.logical_and(rowselect,arevalid)
    data = data.loc[rowselect]
#    print("b",len(data))
    data = data.loc[data.geom_type != 'Point']
    if len(data.loc[data.geom_type == 'GeometryCollection']) > 0:
        print("###########################")
        print("check the following features")
        print(data.loc[data.geom_type == 'GeometryCollection'])
        print("###########################")

    data = data.loc[data.geom_type != 'GeometryCollection']
    data.sindex
#    print("c",len(data))
    return data

def add_area_in_m2(data,prj_crs,area_col):
    src_src = data.crs
    tost = data.copy()

    tost= data.to_crs(prj_crs)
    tost[area_col] = tost.area

    out= tost.copy(deep=True).to_crs(src_src)

    return out

def add_centroid_in_wgs84(data,colx,coly):
    src_src = data.crs
    tost = data.copy()

    tost= tost.to_crs('EPSG:4326')

    tost[coly] = tost.geometry.centroid.y
    tost[colx] = tost.geometry.centroid.x

    out= tost.copy(deep=True).to_crs(src_src)

    return out


def create_geo_jason_file(Input_Polygon_path):

    if "finalcat_info" not in Input_Polygon_path:
#        print(Input_Polygon_path)
        return

    product_dir = os.path.dirname(Input_Polygon_path)
    Names_in = os.path.basename(Input_Polygon_path).split('_')
    n_charc = len(Names_in)
    version  = Names_in[n_charc - 1][0:4]
    TOLERANCEs = [0.0001,0.0005,0.001,0.005,0.01,0.05]


    head_name_cat = "finalcat_info"
    head_name_riv = "finalcat_info_riv"
    head_name_slake = "sl_connected_lake"
    head_name_nlake = "sl_non_connected_lake"

    Input_file_name = []
    Output_file_name = []
    if 'v' in version:
        Input_file_name = [
                           head_name_cat + "_"+version+'.shp',
                           head_name_riv + "_"+version+'.shp',
                           head_name_slake + "_"+version+'.shp',
                           head_name_nlake + "_"+version+'.shp',
                           ]
        Output_file_name = [
                           head_name_cat + "_"+version+'.geojson',
                           head_name_riv + "_"+version+'.geojson',
                           head_name_slake + "_"+version+'.geojson',
                           head_name_nlake + "_"+version+'.geojson',
                           ]
    else:
        Input_file_name = [
                           head_name_cat +'.shp',
                           head_name_riv +'.shp',
                           head_name_slake +'.shp',
                           head_name_nlake +'.shp',
                           ]
        Output_file_name = [
                           head_name_cat +'.geojson',
                           head_name_riv +'.geojson',
                           head_name_slake +'.geojson',
                           head_name_nlake +'.geojson',
                           ]
    created_jason_files = []
    created_jason_files_lake_riv = []

    for i  in  range(0,len(Input_file_name)):
        input_path = os.path.join(product_dir,Input_file_name[i])
        output_jason_path = os.path.join(product_dir,Output_file_name[i])
        if not os.path.exists(input_path):
            continue
        created_jason_files.append(output_jason_path)

        if 'finalcat_info_riv' in Input_file_name[i] or 'connected_lake' in Input_file_name[i]:
            created_jason_files_lake_riv.append(output_jason_path)

        # reproject to WGS84
        input_pd = geopandas.read_file(input_path)

        input_wgs_84 = input_pd.to_crs('EPSG:4326')



        if 'finalcat_info' in Input_file_name[i] or "finalcat_info_riv" in Input_file_name[i]:
            input_wgs_84['rvhName'] = input_wgs_84['SubId'].astype(int).astype(str)
            for idx in input_wgs_84.index:
                input_wgs_84.loc[idx,'rvhName'] = 'sub' + input_wgs_84.loc[idx,'rvhName']


        input_tojson = input_wgs_84

        for TOLERANCE in TOLERANCEs:
            input_tojson['geometry'] = input_tojson.simplify(TOLERANCE)
            input_tojson.to_file(output_jason_path, driver="GeoJSON")

            json_file_size = os.stat(output_jason_path).st_size/1024/1024 #to MB
            if json_file_size <= 100:
                break

    if len(created_jason_files_lake_riv) > 1 and os.stat(os.path.join(product_dir,Output_file_name[0])).st_size/1024/1024 < 500:
        for i in range(0,len(created_jason_files_lake_riv)):
            injson2 = load(open(created_jason_files_lake_riv[i]))
            if 'finalcat_info_riv' in created_jason_files_lake_riv[i]:
                new_features = []
                for element in injson2["features"]:
                    if element["properties"]["Lake_Cat"] == 0:
                        new_features.append(element)
                injson2["features"] = new_features

            if i == 0:
                output_jason_lake_riv = injson2
            else:
                output_jason_lake_riv['features'] += injson2['features']

        with open(os.path.join(product_dir,'routing_product_lake_river.geojson'), 'w', encoding='utf-8') as f:
            json.dump(output_jason_lake_riv, f, ensure_ascii=False, indent=4)

    return



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3


def define_cat_and_riv_without_merge_lake_cats(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    path_lakefile_in,
    catchment_without_merging_lakes="catchment_without_merging_lakes",
    river_without_merging_lakes="river_without_merging_lakes",
    max_memroy=1024 * 4,
):

    fdr_arcgis = input_geo_names["fdr_arcgis"]
    fdr_grass = input_geo_names["fdr_grass"]
    str_r = input_geo_names["str_r"]
    str_v = input_geo_names["str_v"]
    acc = input_geo_names["acc"]
    mask = input_geo_names["mask"]
    dem = input_geo_names["dem"]
    nfdr_grass = input_geo_names["nfdr_grass"]
    sl_non_connected_lake = input_geo_names["sl_non_connected_lake"]
    pourpoints_add_obs = input_geo_names["pourpoints_add_obs"]
    lake_outflow_pourpoints = input_geo_names["lake_outflow_pourpoints"]

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    grass.run_command(
        "r.to.vect",
        input=pourpoints_add_obs,
        output=pourpoints_add_obs,
        type="point",
        overwrite=True,
        flags="v",
    )

    if path_lakefile_in != "#":
        grass.run_command(
            "r.stream.basins",
            direction=nfdr_grass,
            points=pourpoints_add_obs,
            basins="define_cat_and_riv_1",
            overwrite=True,
            memory=max_memroy,
        )
        # find non connected lake catchments
        lakeid, catid = generate_stats_list_from_grass_raster(
            grass,
            mode=2,
            input_a=lake_outflow_pourpoints,
            input_b="define_cat_and_riv_1",
        )
        lakecat_ids = np.column_stack((lakeid, catid))

        # find all non connected cat ids
        non_connected_lakeids, temp = generate_stats_list_from_grass_raster(
            grass,
            mode=1,
            input_a=sl_non_connected_lake,
        )

        non_cl_lake_catid = lakecat_ids[
            np.isin(lakecat_ids[:, 0], np.array(non_connected_lakeids)), 1
        ]

        # find all catids
        all_catids, temp = generate_stats_list_from_grass_raster(
            grass,
            mode=1,
            input_a="define_cat_and_riv_1",
        )
        all_catids = np.array(all_catids)
        not_non_cl_lake_catids = all_catids[
            np.logical_not(np.isin(all_catids, non_cl_lake_catid))
        ]
        # grass.run_command(
        #     "g.copy", rast=("define_cat_and_riv_1", "non_cl_lake_cat"), overwrite=True
        # )
        non_cl_lake_cat_array = garray.array(mapname="define_cat_and_riv_1")
        if (len(not_non_cl_lake_catids) > 0):
            mask = np.isin(non_cl_lake_cat_array, not_non_cl_lake_catids)
            non_cl_lake_cat_array[mask] = -9999
        temparray = garray.array()
        temparray[:, :] = non_cl_lake_cat_array[:, :]
        temparray.write(mapname="non_cl_lake_cat", overwrite=True)
        grass.run_command("r.null", map="non_cl_lake_cat", setnull=[-9999, 0])
        exp = "%s = int(%s)" % (
            "non_cl_lake_cat",
            "non_cl_lake_cat",
        )
        grass.run_command("r.mapcalc", expression=exp, overwrite=True)

        # grass.run_command(
        #     "r.null",
        #     map="non_cl_lake_cat",
        #     setnull=not_non_cl_lake_catids,
        #     overwrite=True,
        # )

        # define str and cat
        grass.run_command(
            "r.cross",
            input=[str_r, "define_cat_and_riv_1"],
            output="nstr_seg_t",
            flags="z",
            overwrite=True,
        )
        grass.run_command(
            "r.mapcalc", expression="nstr_seg = int(nstr_seg_t) + 1", overwrite=True
        )
        grass.run_command(
            "r.stream.basins",
            direction=nfdr_grass,
            stream="nstr_seg",
            basins="define_cat_and_riv_2",
            overwrite=True,
        )

        grass.run_command(
            "r.cross",
            input=["define_cat_and_riv_2", "non_cl_lake_cat"],
            output=catchment_without_merging_lakes,
            overwrite=True,
        )

        exp = "%s = %s + 1" % (
            catchment_without_merging_lakes,
            catchment_without_merging_lakes,
        )
        grass.run_command("r.mapcalc", expression=exp, overwrite=True)

        exp = "%s = if(isnull(int(%s)),null(),%s)" % (
            river_without_merging_lakes,
            str_r,
            catchment_without_merging_lakes,
        )
        grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    else:
        grass.run_command(
            "r.stream.basins",
            direction=fdr_grass,
            points=pourpoints_add_obs,
            basins="define_cat_and_riv_1",
            overwrite=True,
            memory=max_memroy,
        )

        grass.run_command(
            "r.cross",
            input=[str_r, "define_cat_and_riv_1"],
            output="nstr_seg_t",
            flags="z",
            overwrite=True,
        )
        grass.run_command(
            "r.mapcalc", expression="nstr_seg = int(nstr_seg_t) + 1", overwrite=True
        )

        grass.run_command(
            "r.stream.basins",
            direction=fdr_grass,
            stream="nstr_seg",
            basins=catchment_without_merging_lakes,
            overwrite=True,
        )

        exp = "%s = if(isnull(int(%s)),null(),%s)" % (
            river_without_merging_lakes,
            str_r,
            catchment_without_merging_lakes,
        )
        grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    PERMANENT.close()
    return



from basinmaker.utilities.utilities import *
from basinmaker.func.pdtable import *


def add_attributes_to_catchments(
    input_geo_names,
    path_bkfwidthdepth="#",
    bkfwd_attributes=[],
    path_landuse="#",
    path_landuse_info="#",
    path_k_c_zone_polygon = '#',
    k_in=-1,
    c_in=-1,
    out_cat_name="catchment_without_merging_lakes",
    out_riv_name="river_without_merging_lakes",
    grassdb="#",
    grass_location="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
    projection="EPSG:3573",
    obs_attributes=[],
    lake_attributes=[],
    outlet_obs_id=1,
    path_sub_reg_outlets_v="#",
    output_folder="#",
):

    """Calculate hydrological paramters

    Calculate hydrological paramters for each subbasin generated by
    "AutomatedWatershedsandLakesFilterToolset". The result generaed
    by this tool can be used as inputs for Define_Final_Catchment
    and other post processing tools

    Parameters
    ----------
    input_geo_names                : dict
        it is a dictionary that list the required input file names,should at
        least indicate the name of following items:
        sl_nonconnect_lake          : raster
            it is a raster represent all non connected lakes that are selected
            by lake area threstholds
        sl_connected_lake           : raster
            it is a raster represent allconnected lakes that are selected
            by lake area threstholds
        river_without_merging_lakes : raster/vector
            it is the updated river segment for each subbasin
        catchment_without_merging_lakes   : raster/vector
            it is a raster represent updated subbasins after adding lake inflow
            and outflow points as new subbasin outlet.
        snapped_obs_points          : raster/vector
            it is a name of the point gis file represent successfully sanpped
            observation points

    path_bkfwidthdepth             : string
        It is a string to indicate the full path of the
        polyline shapefile that having bankfull width and
        depth data
    bkfwd_attributes               :
        the columns names that indicate following items has to be included
        1) column name for the Bankfull width in m;
        2) column name for the Bankfull depth in m;
        3) column name for the annual mean discharge in m3/s;
    path_landuse                   : string
        It is a string to indicate the full path of the landuse data.
        It will be used to estimate the floodplain roughness
        coefficient. Should have the same projection with the DEM data
        in ".tif" format.
    path_landuse_info              : string
        It is a string to indicate the full path of the table in '.csv'
        format.The table describe the floodplain roughness coefficient
        correspond to a given landuse type. The table should have two
        columns: RasterV and MannV. RasterV is the landuse value in the
        landuse raster for each land use type and the MannV is the
        roughness coefficient value for each landuse type.
    gis_platform                   : string
        It is a string indicate with gis platform is used:
        'qgis'                : the basinmaker is running within QGIS
        'arcgis'              : the basinmaker is running within ArcGIS
    lake_attributes                : list
        the columns names that indicate following items has to be included
        1) column name for the unique Id of each lake within the lake polygon shpfile;
        2) column name for type of the lake should be integer;
        3) column name for the volume of the lake in km3;
        4) column name for the average depth of the lake in m;
        5) column name for the area of the lake in km2.
    obs_attributes                 : list
        the columns names that indicate following items has to be included
        1) column name for the unique Id of each observation point;
        2) column name for the unique name of each observation point;
        3) column name for the drainage area of each observation point in km3;
        4) column name for the source of the observation point:
            'CA' for observation in canada;
            'US' for observation in US;
    outlet_obs_id                  : int
        It is one 'Obs_ID' in the provided observation gauge
        shapefile. If it is larger than zero. Subbasins that
        do not drainage to this gauge will be removed from
        delineation result.
    projection                     : string
        It is a string indicate a projected coordinate system,
        which wiil be used to calcuate area, slope and aspect.
    output_folder                  : string
        The path to a folder to save outputs
    path_sub_reg_outlets_v         : string

    Notes
    -------
    Five vector files will be generated by this function. these files
    can be used to define final routing structure by "Define_Final_Catchment"
    or be used as input for other postprocessing tools. All files
    are stored at self.OutputFolder

    catchment_without_merging_lakes.shp             : shapefile
        It is the subbasin polygon before merging lakes catchments and
        need to be processed before used.
    river_without_merging_lakes.shp                 : shapefile
        It is the subbasin river segment before merging lakes catchments and
        need to be processed before used.
    sl_connected_lake.shp                           : shapefile
        It is the connected lake polygon. Connected lakes are lakes that
        are connected by  Path_final_riv.
    sl_non_connected_lake.shp                       : shapefile
        It is the  non connected lake polygon. Connected lakes are lakes
        that are not connected by Path_final_cat_riv or Path_final_riv.
    obs_gauges                                      : shapefile
        It is the point shapefile that represent the observation gauge
        after snap to river network.

    Returns:
    -------
       None

    Examples
    -------

    """

    columns = COLUMN_NAMES_CONSTANT
    coltypes = COLUMN_TYPES_CONSTANT

    # local geo file names
    cat_ply_info = Internal_Constant_Names["cat_ply_info"]
    cat_riv_info = Internal_Constant_Names["cat_riv_info"]
    outlet_pt_info = Internal_Constant_Names["outlet_pt_info"]

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    if gis_platform == "qgis":
        assert (
            grassdb != "#"
        ), "grass database folder is needed, when gis_platform = qgis"
        assert (
            grass_location != "#"
        ), "grass location name is needed, when gis_platform = qgis"
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.addattributes.createattributestemplateqgis import (
            create_catchments_attributes_template_table,
        )
        from basinmaker.addattributes.calculatebasicattributesqgis import (
            calculate_basic_attributes,
        )
        from basinmaker.addattributes.addlakeattributesqgis import add_lake_attributes
        from basinmaker.addattributes.joinpandastoattributesqgis import (
            join_pandas_table_to_vector_attributes,
        )
        from basinmaker.addattributes.exportoutputsqgis import export_files_to_output_folder
        from basinmaker.addattributes.addgaugeattributesqgis import add_gauge_attributes
        from basinmaker.addattributes.calfloodmanningnqgis import calculate_flood_plain_manning_n
        from basinmaker.addattributes.calbkfwidthdepthqgis import (
            calculate_bankfull_width_depth_from_polyline,
        )

        attr_template = create_catchments_attributes_template_table(
            grassdb=grassdb,
            grass_location=grass_location,
            columns=columns,
            input_geo_names=input_geo_names,
        )
        
        attr_basic = calculate_basic_attributes(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            input_geo_names=input_geo_names,
            projection=projection,
            catinfo=attr_template,
            cat_ply_info=cat_ply_info,
            cat_riv_info=cat_riv_info,
            outlet_pt_info=outlet_pt_info,
        )
        input_geo_names["cat_ply_info"] = cat_ply_info
        input_geo_names["cat_riv_info"] = cat_riv_info
        input_geo_names["outlet_pt_info"] = outlet_pt_info
        
        if len(lake_attributes) > 0:
            attr_lake = add_lake_attributes(
                grassdb=grassdb,
                grass_location=grass_location,
                qgis_prefix_path=qgis_prefix_path,
                input_geo_names=input_geo_names,
                lake_attributes=lake_attributes,
                catinfo=attr_basic,
            )
        else:
            attr_lake = attr_basic
        
        if len(obs_attributes) > 0:
            attr_obs = add_gauge_attributes(
                grassdb=grassdb,
                grass_location=grass_location,
                qgis_prefix_path=qgis_prefix_path,
                input_geo_names=input_geo_names,
                obs_attributes=obs_attributes,
                catinfo=attr_lake,
            )
        else:
            attr_obs = attr_lake
        
        if outlet_obs_id > 0:
            attr_select = return_interest_catchments_info(
                catinfo=attr_obs,
                outlet_obs_id=outlet_obs_id,
                path_sub_reg_outlets_v=path_sub_reg_outlets_v,
            )
        else:
            attr_select = attr_obs
        
        if path_landuse != "#":
            attr_landuse = calculate_flood_plain_manning_n(
                grassdb=grassdb,
                grass_location=grass_location,
                qgis_prefix_path=qgis_prefix_path,
                catinfo=attr_select,
                input_geo_names=input_geo_names,
                path_landuse=path_landuse,
                path_landuse_info=path_landuse_info,
            )
        else:
            attr_landuse = attr_select
        
        attr_da = streamorderanddrainagearea(attr_landuse)
         
        if path_bkfwidthdepth != "#" or k_in != -1 or path_k_c_zone_polygon != '#':
            attr_bkf = calculate_bankfull_width_depth_from_polyline(
                grassdb=grassdb,
                grass_location=grass_location,
                qgis_prefix_path=qgis_prefix_path,
                path_bkfwidthdepth=path_bkfwidthdepth,
                path_k_c_zone_polygon = path_k_c_zone_polygon,
                bkfwd_attributes=bkfwd_attributes,
                catinfo=attr_da,
                input_geo_names=input_geo_names,
                k_in=k_in,
                c_in=k_in,
            )
        else:
            attr_bkf = attr_da
        
        attr_ncl = update_non_connected_catchment_info(attr_bkf)
        attr_ncl.loc[attr_ncl['RivLength'] == -1.2345,'RivSlope'] = -1.2345
        attr_ncl.loc[attr_ncl['RivLength'] == -1.2345,'FloodP_n'] = -1.2345
        attr_ncl.loc[attr_ncl['RivLength'] == -1.2345,'Max_DEM'] = -1.2345
        attr_ncl.loc[attr_ncl['RivLength'] == -1.2345,'Min_DEM'] = -1.2345
        attr_ncl.loc[attr_ncl['RivLength'] == -1.2345,'Ch_n'] = -1.2345
        
        join_pandas_table_to_vector_attributes(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            vector_name=cat_ply_info,
            pd_table=attr_ncl,
            column_types=coltypes,
            columns_names=columns,
        )
        
        join_pandas_table_to_vector_attributes(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            vector_name=cat_riv_info,
            pd_table=attr_ncl,
            column_types=coltypes,
            columns_names=columns,
        )

        export_files_to_output_folder(
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            input_geo_names=input_geo_names,
            output_riv=out_riv_name,
            output_cat=out_cat_name,
            output_folder=output_folder,
            obs_attributes = obs_attributes,
        )



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def define_project_extent_using_hybasin_ply(
    grassdb,
    grass_location,
    qgis_prefix_path,
    path_dem_in,
    buffer_distance,
    hybasin_ply,
    down_hybasin_id,
    up_hybasin_id,
    mask="MASK",
    dem="dem",
):

    """Define processing extent

    Function that used to define project processing spatial extent (PSE).
    The processing spatial extent is a region where Toolbox will work in. Toolbox
    will not process grids or features outside the processing spatial extent.
    Several options is available here. The PSE can be defined using Hybasin
    product and a hydrobasin ID. All subbasin drainage to that hydrobasin ID
    will be extracted. And the extent of the extracted polygon will be used as PSE

    Parameters
    ----------
    grassdb                           : path (required)
        It is a path to project grass database folder
    grass_location                    : string (required)
        It is a string of grass location name
    qgis_prefix_path                  : string (required)
        It is a string of qgis prefix path
    path_dem_in                      : string (required)
        It is the path to input dem
    buffer_distance                  : float (optional)
        It is a float number to increase the extent of the PSE
        obtained from Hydrobasins. It is needed when input DEM is not from
        HydroSHEDS. Then the extent of the watershed will be different
        with PSE defined by HydroBASINS.
    hybasin_ply                      : string (optional)
        It is a path to hydrobasin routing product, If it is provided, the
        PSE will be based on the OutHyID and OutHyID2 and
        this HydroBASINS routing product.
    down_hybasin_id                  : int (optional)
        It is a HydroBASINS subbasin ID, which should be the ID of the most
        downstream subbasin in the region of interest.
    up_hybasin_id                    : int (optional)
        It is a HydroBASINS subbasin ID, which should be the ID of the most
        upstream subbasin in the region of interest, normally do not needed.
    mask                             : string (optional)
        It is a output mask name, which will stored in grass_location in both
        vector and raster format
    dem                              : string (optional)
        It is a output dem raster name, which will be stored in grass_location

    Notes
    -------
    Outputs are following files

    MASK                   : raster
        it is a mask raster stored in grass database, which indicate
        the PSE. The grass database is located at
        os.path.join(grassdb, grass_location)
    dem                   : raster
        it is a dem raster stored in grass database, which is
        has the same extent with MASK. The grass database is located at
        os.path.join(grassdb, grass_location)

    Returns:
    -------
       None

    Examples
    -------
    """

    print("mask region:   using hybasin polygon ")

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    r_dem_layer = qgis_raster_read_raster(
        processing, path_dem_in
    )  ### load DEM raster as a  QGIS raster object to obtain attribute
    cellSize, SpRef_in = qgis_raster_return_raster_properties(
        processing, r_dem_layer
    )  ### Get Raster cell size

    hyshdinfo = Dbf_To_Dataframe(hybasin_ply)
    routing_info = hyshdinfo[["HYBAS_ID", "NEXT_DOWN"]].astype("float").values

    # obtain sub id of subbasins between OutHyID and OutHyID2 in the routing
    # network
    HydroBasins = Return_SubIds_Between_Two_Subbasins_In_Rouing_Network(
        routing_info, down_hybasin_id, up_hybasin_id
    )

    # extract subbasins from hydrobasin product
    Selectfeatureattributes(
        processing,
        Input=hybasin_ply,
        Output=os.path.join(grassdb, mask + "_hy.shp"),
        Attri_NM="HYBAS_ID",
        Values=HydroBasins,
    )

    print("Mask Region:   Using buffered hydroBasin product polygons ")

    # dissolve, buffer and reproject the extracted hydrobasin product
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + "_hy.shp"),
        FIELD="MAIN_BAS",
        OUTPUT=os.path.join(grassdb, mask + "_1hy.shp"),
    )
    qgis_vector_buffer(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + "_1hy.shp"),
        Buffer_Distance=buffer_distance,
        OUTPUT=os.path.join(grassdb, mask + "_2hy.shp"),
    )
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + "_2hy.shp"),
        FIELD="MAIN_BAS",
        OUTPUT=os.path.join(grassdb, mask + "_3hy.shp"),
    )
    qgis_vector_reproject_layers(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + "_3hy.shp"),
        TARGET_CRS=SpRef_in,
        OUTPUT=os.path.join(grassdb, mask + ".shp"),
    )

    # clip raster layer with this mask
    qgis_raster_clip_raster_by_mask(
        processing,
        Input=path_dem_in,
        MASK=os.path.join(grassdb, mask + ".shp"),
        TARGET_CRS=SpRef_in,
        Output=os.path.join(grassdb, mask + ".tif"),
    )

    # use clipped DEM to great a grass work enviroment
    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    # open/create a grass location
    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT_temp = Session()
    PERMANENT_temp.open(
        gisdb=grassdb,
        location=grass_location + "t1",
        create_opts="EPSG:4326",
    )

    # import clipped dem to target location
    grass_raster_r_in_gdal(
        grass=grass,
        raster_path=os.path.join(grassdb, mask + ".tif"),
        output_nm=dem,
        location=grass_location,
    )
    PERMANENT_temp.close()

    # Define mask and processing region for grass working enviroments
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    # define mask of current working enviroments
    grass_raster_r_mask(grass, dem)
    # define processing extent of the current working enviroment
    grass_raster_g_region(grass, dem)

    PERMANENT.close()
    Qgs.exit()



import numpy as np
import sys
import os
import csv
import tempfile 
import copy 
import pandas as pd
from arcgis.features import GeoAccessor, GeoSeriesAccessor
import arcpy
from arcpy import env
from arcpy.sa import *


import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from basinmaker.func.arcgis import *
from basinmaker.func.pdtable import *
arcpy.env.overwriteOutput = True
arcpy.CheckOutExtension("Spatial")

def combine_catchments_covered_by_the_same_lake_arcgis(
    Routing_Product_Folder
):
    """Define final lake river routing structure

    Generate the final lake river routing structure by merging subbasin
    polygons that are covered by the same lake.
    The input are the catchment polygons and river segements
    before merging for lakes. The input files can be output of
    any of following functions:
    SelectLakes, Select_Routing_product_based_SubId,
    Customize_Routing_Topology,RoutingNetworkTopologyUpdateToolset_riv
    The result is the final catchment polygon that ready to be used for
    hydrological modeling

    Parameters
    ----------
    OutputFolder                   : string
        Folder name that stores generated extracted routing product
    Path_final_riv_ply             : string
        Path to the catchment polygon which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.
        routing product and can be directly used.
    Path_final_riv                 : string
        Path to the river polyline which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.

    Notes
    -------
    This function has no return values, instead will generate following
    files. They are catchment polygons and river polylines that can be
    used for hydrological modeling.
    os.path.join(OutputFolder,'finalcat_info.shp')
    os.path.join(OutputFolder,'finalcat_info_riv.shp')

    Returns:
    -------
    None

    Examples
    -------

    """


    sub_colnm = "SubId"

    Path_Catchment_Polygon="#"
    Path_River_Polyline="#"
    Path_Con_Lake_ply="#"
    Path_NonCon_Lake_ply="#"
    Path_obs_gauge_point="#"
    Path_final_cat_ply="#"
    Path_final_cat_riv="#"

    ##define input files from routing prodcut 
    for file in os.listdir(Routing_Product_Folder):
        if file.endswith(".shp"):
            if 'catchment_without_merging_lakes' in file:
                Path_Catchment_Polygon = os.path.join(Routing_Product_Folder, file)
            if 'river_without_merging_lakes' in file:
                Path_River_Polyline = os.path.join(Routing_Product_Folder, file)
            if 'sl_connected_lake' in file:
                Path_Con_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'sl_non_connected_lake' in file:
                Path_NonCon_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'obs_gauges' in file:
                Path_obs_gauge_point = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info' in file:
                Path_final_cat_ply = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info_riv' in file:
                Path_final_cat_riv = os.path.join(Routing_Product_Folder, file)                

    if Path_Catchment_Polygon == '#' or  Path_River_Polyline =='#':
        print("Invalid routing product folder ")


    OutputFolder = Routing_Product_Folder


    Path_final_rivply = Path_Catchment_Polygon
    Path_final_riv = Path_River_Polyline

    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)
        
    tempfolder = os.path.join(
        tempfile.gettempdir(), "basinmaker_" + str(np.random.randint(1, 10000 + 1))
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)
    arcpy.env.workspace = tempfolder

    ### create a copy of shapfiles in temp folder
    Path_Temp_final_rviply = os.path.join(OutputFolder,"temp_finalriv_ply" + str(np.random.randint(1, 10000 + 1)) + ".shp")
    
    Path_Temp_final_rvi = os.path.join(OutputFolder,"temp_finalriv_riv" + str(np.random.randint(1, 10000 + 1)) + ".shp")
    
    ### read riv ply info
    ### read attribute table
    finalrivply_info = pd.DataFrame.spatial.from_featureclass(Path_final_rivply)
    # change attribute table for lake covered catchments,
    finalrivply_info['SubId'] = finalrivply_info['SubId'].astype('int32')
    finalrivply_info['DowSubId'] = finalrivply_info['DowSubId'].astype('int32')
    finalrivply_info['HyLakeId'] = finalrivply_info['HyLakeId'].astype('int32')
#    finalrivply_info['DrainArea'] = finalrivply_info['DrainArea'].astype('float')

    mapoldnew_info = change_attribute_values_for_catchments_covered_by_same_lake(
        finalrivply_info
    )

    mapoldnew_info = remove_possible_small_subbasins(mapoldnew_info = mapoldnew_info, area_thresthold = 10*30*30/1000/1000)
    # update topology for new attribute table
    
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'RivLength'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'RivSlope'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'FloodP_n'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'Ch_n'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'Max_DEM'] = -1.2345
    mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'Min_DEM'] = -1.2345
    
    if 'DA_Chn_L' in mapoldnew_info.columns:
        mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'RivLength'] = -1.2345
        mapoldnew_info.loc[mapoldnew_info['Lake_Cat'] > 0,'RivLength'] = -1.2345
            
    mapoldnew_info = update_topology(mapoldnew_info, UpdateStreamorder=-1)    
    
    mapoldnew_info['DowSubId'] = mapoldnew_info['DowSubId'].astype('int32')
    
    if len(os.path.basename(Path_Catchment_Polygon).split('_')) == 5:
        cat_name = "finalcat_info_"+os.path.basename(Path_Catchment_Polygon).split('_')[4]
        riv_name = "finalcat_info_riv_"+os.path.basename(Path_Catchment_Polygon).split('_')[4]
    else:
        cat_name =  "finalcat_info.shp"
        riv_name =  "finalcat_info_riv.shp"
              
    save_modified_attributes_to_outputs(
        mapoldnew_info=mapoldnew_info,
        tempfolder=tempfolder,
        OutputFolder=OutputFolder,
        cat_name=cat_name,
        riv_name =riv_name,
        Path_final_riv = Path_final_riv,
    )
    return 
